import json
import os
import locale
import gettext
from typing import Union, Dict
from eduvpn.settings import COUNTRY, LANGUAGE, COUNTRY_MAP
from eduvpn.utils import get_logger

logger = get_logger(__name__)

country_mapping = None


def init(lets_connect: bool, prefix: str):
    """
    Init locale and gettext, returns text domain
    """
    domain = 'LetConnect' if lets_connect else 'eduVPN'
    directory = os.path.join(prefix, 'share/locale')

    locale.setlocale(locale.LC_ALL, '')
    locale.bindtextdomain(domain, directory)  # type: ignore
    locale.textdomain(domain)  # type: ignore
    gettext.bindtextdomain(domain, directory)
    gettext.textdomain(domain)

    return domain


def country() -> str:
    try:
        return locale.getlocale()[0].replace('_', '-')
    except Exception:
        return COUNTRY


def language() -> str:
    try:
        return locale.getlocale()[0].split('_')[0]
    except Exception:
        return LANGUAGE


def extract_translation(d: Union[str, Dict[str, str]]):
    if isinstance(d, dict):
        for m in [country(), language(), 'en-US', 'en']:
            try:
                return d[m]
            except KeyError:
                continue
        return list(d.values())[0]  # otherwise just return first in list
    else:
        return d


def retrieve_country_name(country_code: str) -> str:
    country_map = _read_country_map()
    loc = locale.getlocale()
    if loc[0] is None:
        prefix = 'en'
    else:
        prefix = loc[0][:2]
    if country_code in country_map:
        code = country_map[country_code]
        if prefix in code:
            return code[prefix]
    return country_code

# def get_token(auth_url: str) -> Optional[Tuple[OAuth2Token, str, str]]:
#     """
#     Return the metadata from storage
#     """
#     storage = _read_tokens()
#     if auth_url in storage:
#         v = storage[auth_url]
#         return OAuth2Token(v['token']), v['token_endpoint'], v['authorization_endpoint']
#     else:
#         return None


def _read_country_map() -> dict:
    """
    Read the storage from disk, returns an empty dict in case of failure.
    """
    global country_mapping

    if country_mapping is not None:
        return country_mapping

    if COUNTRY_MAP.exists():
        try:
            with open(COUNTRY_MAP, 'r') as f:
                country_mapping = json.load(f)
                return country_mapping
        except Exception as e:
            logger.error(f"Error reading country map: {e}")
    return {}



from eduvpn.utils import logger
import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
from gi.repository import Gtk, GObject


# ui thread
def error_helper(parent: GObject,  # type: ignore
                 msg_big: str,
                 msg_small: str) -> None:
    """
    Shows a GTK error message dialog.
    args:
        parent (GObject): A GTK Window
        msg_big (str): the big string
        msg_small (str): the small string
    """
    logger.error(f"{msg_big}: {msg_small}")
    error_dialog = Gtk.MessageDialog(  # type: ignore
        parent,
        0,
        Gtk.MessageType.ERROR,  # type: ignore
        Gtk.ButtonsType.OK,  # type: ignore
        str(msg_big),
    )
    error_dialog.format_secondary_text(str(msg_small))  # type: ignore
    error_dialog.run()  # type: ignore
    error_dialog.hide()  # type: ignore


def show_ui_component(builder, component: str, show: bool):
    """
    Set the visibility of a UI component.
    """
    component = builder.get_object(component)
    if show:
        component.show()  # type: ignore
    else:
        component.hide()  # type: ignore


def link_markup(link: str) -> str:
    try:
        scheme, rest = link.split(':', 1)
        if rest.startswith('//'):
            rest = rest[2:]
    except ValueError:
        return link
    else:
        return f'<a href="{link}">{rest}</a>'



"""
This module contains code to maintain a simple metadata storage in ~/.config/eduvpn/
"""
from typing import Optional, Tuple, List
from enum import Enum
from os import PathLike
from datetime import datetime
import json
from oauthlib.oauth2.rfc6749.tokens import OAuth2Token
from eduvpn.settings import CONFIG_PREFIX
from eduvpn.utils import get_logger

logger = get_logger(__name__)

_metadata_path = CONFIG_PREFIX / "metadata.json"


class ConnectionType(str, Enum):
    INSTITUTE = "INSTITUTE",
    SECURE = "SECURE",
    OTHER = "OTHER"


def get_all_metadatas() -> dict:
    """
    Read the metadata from disk, returns an empty dict in case of failure.
    """
    if _metadata_path.exists():
        try:
            with open(_metadata_path, 'r') as f:
                return json.load(f)
        except Exception as e:
            logger.error(f"Error reading metadatas {_metadata_path}: {e}")
    return {}


def _write_metadatas(storage: dict) -> None:
    """
    Write the storage to disk.
    """
    try:
        dump = json.dumps(storage)
        _metadata_path.parent.mkdir(parents=True, exist_ok=True)
        with open(_metadata_path, 'w') as f:
            f.write(dump)
    except Exception as e:
        logger.error(f"Error writing metadatas: {e}")


def _get_setting(what: str) -> Optional[str]:
    p = (CONFIG_PREFIX / what).expanduser()
    if p.exists():
        return open(p, 'r').read().strip()
    else:
        return None


def _set_setting(what: str, value: str):
    p = (CONFIG_PREFIX / what).expanduser()
    p.parent.mkdir(parents=True, exist_ok=True)
    with open(p, 'w') as f:
        f.write(value)


Metadata = Tuple[OAuth2Token, str, str, str, str, str, str, str, str, Optional[datetime], Optional[datetime]]


def get_current_metadata(auth_url: str) -> Optional[Metadata]:
    """
    Return the metadata for current connection from storage.
    """
    storage = get_all_metadatas()
    if auth_url in storage:
        v = storage[auth_url]
        created = v.get('certificate_created')
        if created is not None:
            created = datetime.fromisoformat(created)
        expiry = v.get('certificate_expiry')
        if expiry is not None:
            expiry = datetime.fromisoformat(expiry)
        return (
            OAuth2Token(v['token']),
            v['token_endpoint'],
            v['authorization_endpoint'],
            v['api_url'],
            v['display_name'],
            v['support_contact'],
            v['profile_id'],
            v['con_type'],
            v['country_id'],
            created,
            expiry,
        )
    else:
        return None


def set_metadata(
        auth_url: str,
        token: OAuth2Token,
        token_endpoint: str,
        authorization_endpoint: str,
        api_url: str,
        display_name: str,
        support_contact: List[str],
        profile_id: str,
        con_type: str,
        country_id: Optional[str],
        certificate_created: Optional[datetime] = None,
        certificate_expiry: Optional[datetime] = None,
) -> None:
    """
    Set a configuration profile in storage
    """
    storage = get_all_metadatas()
    if certificate_created is None:
        created_str = None
    else:
        created_str = certificate_created.isoformat()
    if certificate_expiry is None:
        expiry_str = None
    else:
        expiry_str = certificate_expiry.isoformat()
    storage[auth_url] = {
        'token': token,
        'api_base_uri': auth_url,
        'token_endpoint': token_endpoint,
        'authorization_endpoint': authorization_endpoint,
        'api_url': api_url,
        'display_name': display_name,
        'support_contact': support_contact,
        'profile_id': profile_id,
        'con_type': con_type,
        'country_id': country_id,
        'certificate_created': created_str,
        'certificate_expiry': expiry_str,
    }
    _write_metadatas(storage)


def del_metadata(auth_url: str) -> None:
    """
    Remove a metadata from the metadata storage
    """
    storage = get_all_metadatas()
    if auth_url in storage:
        storage.pop(auth_url)
        _write_metadatas(storage)


def get_uuid() -> Optional[str]:
    """
    Read the UUID of the last generated eduVPN Network Manager connection.
    """
    return _get_setting("uuid")


def set_uuid(uuid: str):
    """
    Write the eduVPN network manager connection UUID to disk.
    """
    return _set_setting("uuid", uuid)


def get_auth_url() -> Optional[str]:
    """
    Read the auth_url of the current eduVPN Network Manager connection.
    """
    return _get_setting("auth_url")


def set_auth_url(auth_url: str):
    """
    Write the eduVPN network manager active auth_url to disk.
    """
    return _set_setting("auth_url", auth_url)


def write_config(config: str, private_key: str, certificate: str, target: PathLike):
    """
    Write the configuration to target.
    """
    logger.info(f"Writing configuration to {target}")
    with open(target, mode='w+t') as f:
        f.writelines(config)
        f.writelines(f"\n<key>\n{private_key}\n</key>\n")
        f.writelines(f"\n<cert>\n{certificate}\n</cert>\n")


def get_storage(check=False) -> Tuple[Optional[str], Optional[str], Optional[Metadata]]:
    """

    Args:
        check: fail if item not found

    Returns:
        uuid, auth_url, api_url, metadata, profile
    """
    uuid = get_uuid()
    if not uuid and check:
        raise Exception("no eduVPN uuid stored (yet)")

    auth_url = get_auth_url()
    if not auth_url and check:
        raise Exception("no eduVPN auth_url stored (yet)")

    if auth_url:
        metadata = get_current_metadata(auth_url)
        if not metadata and check:
            raise Exception(f"no eduVPN metadata for {auth_url} stored (yet)")
    else:
        metadata = None

    return uuid, auth_url, metadata


def update_token(token: OAuth2Token):
    """
    In case of a token refresh only the new token needs to be written to storage.
    """
    auth_url = get_auth_url()
    logger.info(f"updating token for {auth_url}")
    metadatas = get_all_metadatas()
    if 'auth_url' in metadatas:
        metadatas[auth_url]['token'] = token
        _write_metadatas(metadatas)



from logging import getLogger
from requests_oauthlib import OAuth2Session
from typing import List, Optional

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('NM', '1.0')
from gi.repository import NM  # type: ignore

from eduvpn.remote import list_organisations, list_servers
from eduvpn.settings import SERVER_URI, ORGANISATION_URI
from eduvpn.ui.vpn_connection import VpnConnection

logger = getLogger(__name__)


class BackendData:
    def __init__(self, lets_connect: bool = False):
        self.uuid: Optional[str] = ""
        if lets_connect:
            self.servers = []
        else:
            self.servers = list_servers(SERVER_URI)
        self.secure_internet = [s for s in self.servers if s['server_type'] == 'secure_internet']
        self.institute_access = [s for s in self.servers if s['server_type'] == 'institute_access']

        if lets_connect:
            self.organisations = []
        else:
            self.organisations = list_organisations(ORGANISATION_URI)

        self.profiles: dict = {}
        self.locations: list = []
        self.secure_internet_home: str = ""
        self.oauth: OAuth2Session
        self.new_server_name: str = ""
        self.new_server_image = None
        self.new_support_contact: List[str] = []
        self.connection_state = NM.VpnConnectionState.UNKNOWN
        self.vpn_connection: VpnConnection = VpnConnection()



from typing import Optional
import logging
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.errors import (
    InvalidGrantError as InvalidOauthGrantError)
from .. import storage
from .. import nm
from .. import actions
from .. import remote
from .. import crypto
from ..oauth2 import OAuthWebServer
from ..app import Application
from ..server import (
    AnyServer, ConfiguredServer, InstituteAccessServer,
    SecureInternetServer, OrganisationServer, CustomServer,
    SecureInternetLocation, Profile)
from ..utils import run_in_background_thread


logger = logging.getLogger(__name__)


@run_in_background_thread('start-oauth-web-server')
def on_setup_oauth(app: Application, server: AnyServer):
    try:
        server_info = app.server_db.get_server_info(server)
    except Exception as e:
        logger.error("error getting server info", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return

    def oauth_token_callback(oauth_session: Optional[OAuth2Session]):
        if oauth_session:
            app.interface_transition_threadsafe('oauth_setup_success', oauth_session)
        else:
            # The process has been canceled by the user through the app,
            # which already schedules the state transition so we don't need
            # to do that here.
            pass

    web_server = OAuthWebServer.start(
        server_info.token_endpoint,
        server_info.auth_endpoint,
        oauth_token_callback,
    )
    app.interface_transition_threadsafe('ready_for_oauth_setup', web_server)


@run_in_background_thread('oauth-refresh')
def on_refresh_oauth_token(app: Application,
                           server: AnyServer,
                           oauth_session: OAuth2Session):
    try:
        server_info = app.server_db.get_server_info(server)
    except Exception as e:
        logger.error("error getting server info", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return
    try:
        oauth_session.refresh_token(token_url=server_info.token_endpoint)
    except InvalidOauthGrantError as e:
        logger.warning(f'Error refreshing OAuth token: {e}')
        app.interface_transition_threadsafe('oauth_refresh_failed')
    else:
        app.interface_transition_threadsafe('oauth_refresh_success')


@run_in_background_thread('load-server-info')
def on_start_connection(app: Application,
                        server: AnyServer,
                        oauth_session: OAuth2Session,
                        location: Optional[SecureInternetServer] = None):
    try:
        server_info = app.server_db.get_server_info(server)
    except Exception as e:
        logger.error("error getting server info", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return
    api_url = server_info.api_base_uri
    profile_server: ConfiguredServer

    if isinstance(server, OrganisationServer):
        if not location:
            locations = [server for server in app.server_db.all()
                         if isinstance(server, SecureInternetServer)]
            app.interface_transition_threadsafe(
                'choose_secure_internet_location', server, oauth_session, locations)
            return
        else:
            try:
                location_info = app.server_db.get_server_info(location)
            except Exception as e:
                logger.error("error getting server info", exc_info=True)
                enter_error_state_threadsafe(app, e)
                return
            api_url = location_info.api_base_uri
            profile_server = SecureInternetLocation(server, location)
    else:
        profile_server = server

    profiles = [Profile(**data) for data in remote.list_profiles(oauth_session, api_url)]
    app.interface_transition_threadsafe(
        'choose_profile', profile_server, oauth_session, profiles)


@run_in_background_thread('configure-connection')
def on_chosen_profile(app: Application,
                      server: AnyServer,
                      oauth_session: OAuth2Session,
                      profile: Profile):
    country_code: Optional[str]
    if isinstance(server, SecureInternetLocation):
        try:
            server_info = app.server_db.get_server_info(server.server)
            location_info = app.server_db.get_server_info(server.location)
        except Exception as e:
            logger.error("error getting server info", exc_info=True)
            enter_error_state_threadsafe(app, e)
            return
        auth_url = server.server.oauth_login_url
        api_url = location_info.api_base_uri
        country_code = server.location.country_code
        con_type = storage.ConnectionType.SECURE
        display_name = str(server.server)
        support_contact = server.location.support_contact
    else:
        try:
            server_info = app.server_db.get_server_info(server)
        except Exception as e:
            logger.error("error getting server info", exc_info=True)
            enter_error_state_threadsafe(app, e)
            return
        api_url = server_info.api_base_uri
        auth_url = server.oauth_login_url
        country_code = None
        display_name = str(server)
        if isinstance(server, InstituteAccessServer):
            con_type = storage.ConnectionType.INSTITUTE
            support_contact = server.support_contact
        elif isinstance(server, CustomServer):
            con_type = storage.ConnectionType.OTHER
            support_contact = []
        else:
            raise TypeError(server)

    try:
        config, private_key, certificate = actions.get_config_and_keycert(
            oauth_session, api_url, profile.id)
    except Exception as e:
        logger.error("error getting config and keycert", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return
    validity = crypto.get_certificate_validity(certificate)
    if validity is None:
        validity_start = validity_end = None
    else:
        validity_start, validity_end = validity.start, validity.end
    storage.set_metadata(
        auth_url, oauth_session.token, server_info.token_endpoint,
        server_info.auth_endpoint, api_url, display_name,
        support_contact, profile.id, con_type, country_code,
        validity_start, validity_end)
    storage.set_auth_url(auth_url)

    def finished_saving_config_callback(result):
        logger.info(f"Finished saving network manager config: {result}")
        app.interface_transition('finished_configuring_connection', validity)
        app.current_network_uuid = storage.get_uuid()
        assert app.current_network_uuid is not None
        app.network_transition('start_new_connection', server)

    @app.make_func_threadsafe
    def save_connection():
        nm.save_connection(nm.get_client(), config, private_key, certificate,
                           callback=finished_saving_config_callback)

    save_connection()


def enter_error_state(app: Application, error: Exception):
    app.interface_transition('encountered_exception', error)


def enter_error_state_threadsafe(app: Application, error: Exception):
    app.make_func_threadsafe(enter_error_state)(app, error)



from typing import Union, Optional, Iterable, List, Dict
import os
from eduvpn import remote
from eduvpn import storage
from eduvpn.i18n import extract_translation, retrieve_country_name
from eduvpn.settings import SERVER_URI, ORGANISATION_URI, FLAG_PREFIX
from .utils import custom_server_oauth_url


class InstituteAccessServer:
    """
    A record from: https://disco.eduvpn.org/v2/server_list.json
    where: server_type == "institute_access"
    """

    def __init__(self,
                 base_url: str,
                 display_name: Union[str, Dict[str, str]],
                 support_contact: List[str] = [],
                 keyword_list: Optional[Union[str, List[str]]] = None):
        self.base_url = base_url
        self.display_name = display_name
        self.support_contact = support_contact
        if keyword_list is not None:
            if isinstance(keyword_list, str):
                keyword_list = [keyword_list]
            elif not isinstance(keyword_list, list):
                raise TypeError(keyword_list)
        self.keyword_list = keyword_list

    def __str__(self):
        return extract_translation(self.display_name)

    def __repr__(self):
        return f"<InstituteAccessServer {str(self)!r}>"

    @property
    def search_texts(self):
        texts = [str(self)]
        if self.keyword_list:
            texts.extend(self.keyword_list)
        return texts

    @property
    def oauth_login_url(self):
        return self.base_url


class SecureInternetServer:
    """
    A record from: https://disco.eduvpn.org/v2/server_list.json
    where: server_type == "secure_internet"
    """

    def __init__(self,
                 base_url: str,
                 public_key_list: List[str],
                 country_code: str,
                 support_contact: List[str] = [],
                 authentication_url_template: Optional[str] = None):
        self.base_url = base_url
        self.public_key_list = public_key_list
        self.support_contact = support_contact
        self.country_code = country_code
        self.authentication_url_template = authentication_url_template

    def __str__(self):
        return f"{self.country_name} @ {self.base_url}"

    def __repr__(self):
        return f"<SecureInternetServer {str(self)!r}>"

    @property
    def country_name(self) -> str:
        return retrieve_country_name(self.country_code)

    @property
    def flag_path(self) -> Optional[str]:
        path = f'{FLAG_PREFIX}{self.country_code}@1,5x.png'
        if os.path.exists(path):
            return path
        else:
            return None

    @property
    def oauth_login_url(self):
        return self.base_url


class OrganisationServer:
    """
    A record from: https://disco.eduvpn.org/v2/organization_list.json
    """

    def __init__(self,
                 secure_internet_home: str,
                 display_name: Union[str, Dict[str, str]],
                 org_id: str,
                 keyword_list: Dict[str, str] = {}):
        self.secure_internet_home = secure_internet_home
        self.display_name = display_name
        self.org_id = org_id
        self.keyword_list = keyword_list

    def __str__(self):
        return extract_translation(self.display_name)

    def __repr__(self):
        return f"<OrganisationServer {str(self)!r}>"

    @property
    def keyword(self) -> Optional[str]:
        if self.keyword_list:
            return extract_translation(self.keyword_list)
        return None

    @property
    def search_texts(self):
        texts = [str(self)]
        if self.keyword:
            texts.append(self.keyword)
        return texts

    @property
    def oauth_login_url(self):
        return self.secure_internet_home


class CustomServer:
    """
    A server defined by the user.
    """

    def __init__(self, address: str):
        self.address = address

    def __str__(self):
        return self.address

    def __repr__(self):
        return f"<CustomServer {str(self)!r}>"

    @property
    def oauth_login_url(self):
        return custom_server_oauth_url(self.address)


class ServerInfo:
    def __init__(self, api_base_uri, token_endpoint, auth_endpoint):
        self.api_base_uri = api_base_uri
        self.token_endpoint = token_endpoint
        self.auth_endpoint = auth_endpoint


class Profile:
    def __init__(self,
                 profile_id: str,
                 display_name: str,
                 two_factor: bool,
                 default_gateway: bool):
        self.profile_id = profile_id
        self.display_name = display_name
        self.two_factor = two_factor
        self.default_gateway = default_gateway

    @property
    def id(self):
        return self.profile_id

    def __str__(self):
        return self.display_name

    def __repr__(self):
        return f"<Profile id={self.id!r} {str(self)!r}>"


class SecureInternetLocation:
    def __init__(self,
                 server: OrganisationServer,
                 location: SecureInternetServer):
        self.server = server
        self.location = location

    def __str__(self):
        return self.location.country_name

    def __repr__(self):
        return f"<SecureInternetLocation {self.server!r} {self.location!r}>"

    @property
    def image_path(self) -> Optional[str]:
        return self.location.flag_path

    @property
    def support_contact(self) -> List[str]:
        return self.location.support_contact

    @property
    def oauth_login_url(self):
        assert self.server.oauth_login_url == self.location.oauth_login_url
        return self.server.oauth_login_url


# typing aliases
PredefinedServer = Union[
    InstituteAccessServer,
    OrganisationServer,
    CustomServer,
]
ConfiguredServer = Union[
    InstituteAccessServer,
    SecureInternetLocation,
    CustomServer,
]
AnyServer = Union[
    PredefinedServer,
    ConfiguredServer,
]


def is_search_match(server: PredefinedServer, query: str) -> bool:
    if hasattr(server, 'search_texts'):
        return any(query.lower() in search_text.lower()
                   for search_text
                   in server.search_texts)  # type: ignore
    else:
        return False


class ServerDatabase:
    def __init__(self):
        # TODO load the servers from a cache
        self.servers = []
        self.is_loaded = False

    def update(self):
        """
        Download the list of institute and secure internet servers,
        and update this database.

        This method must be thread-safe.
        """
        new_servers = []
        server_list = remote.list_servers(SERVER_URI)
        organisation_list = remote.list_organisations(ORGANISATION_URI)
        for server_data in server_list:
            server_type = server_data.pop('server_type')
            if server_type == 'institute_access':
                server = InstituteAccessServer(**server_data)
                new_servers.append(server)
            elif server_type == 'secure_internet':
                server = SecureInternetServer(**server_data)
                new_servers.append(server)
            else:
                raise ValueError(server_type, server_data)
        for organisation_data in organisation_list:
            server = OrganisationServer(**organisation_data)
            new_servers.append(server)
        # Atomic update of server map.
        # TODO keep custom other servers
        self.servers = new_servers
        self.is_loaded = True

    def all_configured(self) -> Iterable[ConfiguredServer]:
        "Return all configured servers."
        for key, data in storage.get_all_metadatas().items():
            if data['con_type'] == storage.ConnectionType.INSTITUTE:
                yield InstituteAccessServer(
                    base_url=data['api_base_uri'],
                    display_name=data['display_name'],
                    support_contact=data['support_contact'],
                    keyword_list=None,  # TODO
                )
            elif data['con_type'] == storage.ConnectionType.SECURE:
                secure_internet = SecureInternetServer(
                    base_url=data['api_base_uri'],
                    public_key_list=[],  # TODO
                    country_code=data['country_id'],
                    support_contact=data['support_contact'],
                    authentication_url_template=None,  # TODO
                )
                organisation = OrganisationServer(
                    secure_internet_home=data['api_base_uri'],
                    display_name=data['display_name'],
                    org_id='',  # TODO
                    keyword_list={},
                )
                yield SecureInternetLocation(organisation, secure_internet)
            elif data['con_type'] == storage.ConnectionType.OTHER:
                yield CustomServer(data['api_base_uri'])
            else:
                raise ValueError(data)

    def get_single_configured(self) -> Optional[ConfiguredServer]:
        auth_url = storage.get_auth_url()
        if auth_url is not None:
            for server in self.all_configured():
                if server.oauth_login_url == auth_url:
                    return server
        return None

    def all(self) -> Iterable[PredefinedServer]:
        "Return all servers."
        return iter(self.servers)

    def get_secure_internet_server(self, base_url: str) -> Optional[SecureInternetServer]:
        for server in self.all():
            if isinstance(server, SecureInternetServer) and server.base_url == base_url:
                return server
        return None

    def search(self, query: str) -> Iterable[PredefinedServer]:
        "Return all servers that match the search query."
        if query:
            for server in self.all():
                if is_search_match(server, query):
                    yield server
        else:
            yield from self.all()

    def get_server_info(self, server) -> ServerInfo:
        info = remote.get_info(server.oauth_login_url)
        return ServerInfo(*info)



import logging
import time
from functools import lru_cache
from pathlib import Path
from shutil import rmtree
from sys import modules
from tempfile import mkdtemp
from typing import Any, Optional, Tuple, Callable

from eduvpn.storage import set_uuid, get_uuid, write_config

_logger = logging.getLogger(__name__)

try:
    import gi

    gi.require_version('NM', '1.0')
    from gi.repository import NM, GLib  # type: ignore
except (ImportError, ValueError):
    _logger.warning("Network Manager not available")
    NM = None

try:
    import dbus
except ImportError:
    dbus = None


@lru_cache()
def get_client() -> 'NM.Client':
    """
    Create a new client object. We put this here so other modules don't need to import NM
    """
    return NM.Client.new(None)


@lru_cache()
def get_mainloop():
    """
    Create a new client object. We put this here so other modules don't need to import Glib
    """
    return GLib.MainLoop()


def nm_available() -> bool:
    """
    check if Network Manager is available
    """
    if bool('gi.repository.NM' in modules):
        try:
            get_client()
            return True
        except Exception:
            ...
    return False


def get_existing_configuration_uuid() -> Optional[str]:
    uuid = get_uuid()
    if uuid is None:
        return None
    client = get_client()
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        return None
    else:
        return uuid


def nm_ovpn_import(target: Path) -> Optional['NM.Connection']:
    """
    Use the Network Manager VPN config importer to import an OpenVPN configuration file.
    """
    vpn_infos = [i for i in NM.VpnPluginInfo.list_load() if i.get_name() == 'openvpn']

    if len(vpn_infos) != 1:
        raise Exception(f"Expected one openvpn VPN plugins, got: {len(vpn_infos)}")

    conn = vpn_infos[0].load_editor_plugin().import_(str(target))
    conn.normalize()
    return conn


def import_ovpn(config: str, private_key: str, certificate: str) -> 'NM.SimpleConnection':
    """
    Import the OVPN string into Network Manager.
    """
    target_parent = Path(mkdtemp())
    target = target_parent / "eduVPN.ovpn"
    write_config(config, private_key, certificate, target)
    connection = nm_ovpn_import(target)
    rmtree(target_parent)
    return connection


def add_connection_callback(client, result, callback=None):
    new_con = client.add_connection_finish(result)
    set_uuid(uuid=new_con.get_uuid())
    _logger.info(f"Connection added for uuid: {get_uuid()}")
    if callback is not None:
        callback(new_con is not None)


def add_connection(client: 'NM.Client', connection: 'NM.Connection', callback=None):
    _logger.info("Adding new connection")
    client.add_connection_async(connection=connection, save_to_disk=True, callback=add_connection_callback,
                                user_data=callback)


def update_connection_callback(remote_connection, result, callback=None):
    res = remote_connection.commit_changes_finish(result)
    _logger.debug(f"Connection updated for uuid: {get_uuid()}, result: {res}, remote_con: {remote_connection}")
    if callback is not None:
        callback(result)


def update_connection(old_con: 'NM.Connection', new_con: 'NM.Connection', callback=None):
    """
    Update an existing Network Manager connection with the settings from another Network Manager connection
    """
    _logger.info("Updating existing connection with new configuration")

    # Don't attempt to overwrite the uuid,
    # but reuse the one from the previous connection.
    # Refer to issue #269.
    for setting in new_con.get_settings():
        if setting.get_name() == 'connection':
            setting.props.uuid = old_con.get_uuid()
    old_con.replace_settings_from_connection(new_con)
    old_con.commit_changes_async(save_to_disk=True, cancellable=None, callback=update_connection_callback,
                                 user_data=callback)


def save_connection(client: 'NM.Client', config, private_key, certificate, callback=None):
    _logger.info("writing configuration to Network Manager")
    new_con = import_ovpn(config, private_key, certificate)
    uuid = get_uuid()
    if uuid:
        old_con = client.get_connection_by_uuid(uuid)
        if old_con:
            update_connection(old_con, new_con, callback)
            return
    add_connection(client=client, connection=new_con, callback=callback)


def get_cert_key(client: 'NM.Client', uuid: str) -> Tuple[str, str]:
    try:
        connection = client.get_connection_by_uuid(uuid)
        cert_path = connection.get_setting_vpn().get_data_item('cert')
    except Exception:
        _logger.error(f"Can't fetch stored VPN connecton with uuid {uuid}")
        raise IOError("Can't fetch eduVPN profile")

    key_path = connection.get_setting_vpn().get_data_item('key')
    cert = open(cert_path).read()
    key = open(key_path).read()
    return cert, key


def activate_connection(client: 'NM.Client', uuid: str, callback=None):
    con = client.get_connection_by_uuid(uuid)
    _logger.info(f"activate_connection uuid: {uuid} connection: {con}")
    if con is None:
        # Temporary workaround, connection is sometimes created too
        # late while according to the logging the connection is already
        # created. Need to find the correct event to sync on.
        time.sleep(.1)
        GLib.idle_add(lambda: activate_connection(client, uuid, callback))
        return

    def activate_connection_callback(a_client, res, callback=None):
        try:
            result = a_client.activate_connection_finish(res)
        except Exception as e:
            _logger.error(e)
        else:
            _logger.info(F"activate_connection_async result: {result}")
        finally:
            if callback:
                callback()

    client.activate_connection_async(connection=con, callback=activate_connection_callback, user_data=callback)


def deactivate_connection(client: 'NM.Client', uuid: str, callback=None):
    con = client.get_primary_connection()
    _logger.debug(f"deactivate_connection uuid: {uuid} connection: {con}")
    if con:
        active_uuid = con.get_uuid()

        if uuid == active_uuid:
            def on_deactivate_connection(a_client: 'NM.Client', res, callback=None):
                try:
                    result = a_client.deactivate_connection_finish(res)
                except Exception as e:
                    _logger.error(e)
                else:
                    _logger.info(F"deactivate_connection_async result: {result}")
                finally:
                    if callback:
                        callback()

            client.deactivate_connection_async(active=con, callback=on_deactivate_connection, user_data=callback)
    else:
        _logger.info("No active connection to deactivate")


def connection_status(client: 'NM.Client') -> Tuple[Optional[str], Optional['NM.ActiveConnectionState']]:
    con = client.get_primary_connection()
    if type(con) != NM.VpnConnection:
        return None, None
    uuid = con.get_uuid()
    status = con.get_state()
    return uuid, status


def get_vpn_status(client: 'NM.Client') -> Tuple['NM.VpnConnectionState', 'NM.VpnConnectionStateReason']:
    vpns = [a for a in NM.Client.new(None).get_active_connections() if type(a) == NM.VpnConnection]
    if len(vpns) > 1:
        _logger.warning("more than one VPN connection active")
        return NM.VpnConnectionState.UNKNOWN, NM.VpnConnectionStateReason.UNKNOWN
    elif len(vpns) == 0:
        return NM.VpnConnectionState.UNKNOWN, NM.VpnConnectionStateReason.UNKNOWN
    else:
        return vpns[0].get_state(), vpns[0].get_state_reason()


@lru_cache(maxsize=1)
def get_dbus() -> Optional['dbus.SystemBus']:
    """
    Get the DBus system bus.

    None is returned on failure.
    """
    if dbus is None:
        logging.debug("DBus module could not be imported")
        return None
    try:
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        bus = dbus.SystemBus(private=True)
    except Exception:
        logging.debug("Unable to access dbus", exc_info=True)
        return None
    else:
        return bus


def subscribe_to_status_changes(
    callback: Callable[['NM.VpnConnectionState', 'NM.VpnConnectionStateReason'], Any],
) -> bool:
    """
    Subscribe to all network status changes via DBus.

    The callback argument is called with the connection state and reason
    whenever they change.

    False is returned on failure.
    """
    bus = get_dbus()
    if bus is None:
        return False

    def wrapped_callback(state_code: 'dbus.UInt32', reason_code: 'dbus.UInt32'):
        state = NM.VpnConnectionState(state_code)
        reason = NM.VpnConnectionStateReason(reason_code)
        callback(state, reason)

    bus.add_signal_receiver(
        handler_function=wrapped_callback,
        dbus_interface='org.freedesktop.NetworkManager.VPN.Connection',
        signal_name='VpnStateChanged',
    )
    return True


def action_with_mainloop(action: Callable):
    _logger.info("calling action with CLI mainloop")
    main_loop = get_mainloop()

    def quit_loop(*args, **kwargs):
        _logger.info("Quiting main loop, thanks!")
        main_loop.quit()

    action(callback=quit_loop)
    main_loop.run()


def save_connection_with_mainloop(config, private_key, certificate):
    action_with_mainloop(
        action=lambda callback: save_connection(get_client(), config, private_key, certificate, callback))


def activate_connection_with_mainloop(uuid):
    action_with_mainloop(action=lambda callback: activate_connection(get_client(), uuid, callback))


def deactivate_connection_with_mainloop(uuid):
    action_with_mainloop(action=lambda callback: deactivate_connection(get_client(), uuid, callback))



import logging
from .server import ServerDatabase
from . import nm
from . import storage
from .crypto import Validity
from .state_machine import StateMachine, InvalidStateTransition
from .utils import run_in_background_thread


logger = logging.getLogger(__name__)


class Application:
    def __init__(self, make_func_threadsafe):
        self.make_func_threadsafe = make_func_threadsafe
        from .network import InitialNetworkState
        from .interface.state import InitialInterfaceState
        self.network_state_machine = StateMachine(InitialNetworkState())
        self.interface_state_machine = StateMachine(InitialInterfaceState())
        self.server_db = ServerDatabase()
        self.current_network_uuid = None

    def initialize(self):
        self.initialize_network()
        self.initialize_server_db()

    def initialize_network(self):
        """
        Determine the current network state.
        """
        # Check if a previous network configuration exists.
        uuid = nm.get_existing_configuration_uuid()
        kwargs = {}
        if uuid:
            self.current_network_uuid = uuid
            # Check what server corresponds to the configuration.
            server = self.server_db.get_single_configured()
            if server is None:
                # There is a network configuration,
                # but no record of what server corresponds to it.
                transition = 'no_previous_connection_found'
            else:
                status_uuid, status = nm.connection_status(nm.get_client())
                if status in [nm.NM.ActiveConnectionState.ACTIVATED,
                              nm.NM.ActiveConnectionState.ACTIVATING]:
                    assert uuid == status_uuid
                    *_, start, end = storage.get_current_metadata(server.oauth_login_url)
                    if start is not None and end is not None:
                        validity = Validity(start, end)
                    else:
                        validity = None
                    transition = 'found_active_connection'
                    kwargs['server'] = server
                    kwargs['validity'] = validity
                else:
                    transition = 'found_previous_connection'
                    kwargs['server'] = server
        else:
            transition = 'no_previous_connection_found'
        self.network_transition(transition, **kwargs)

        def on_network_update_callback(state, reason):
            network.on_status_update_callback(self, state)

        from . import network
        if not nm.subscribe_to_status_changes(on_network_update_callback):
            logger.warning(
                "unable to subscribe to network updates; "
                "the application may not reflect the current state"
            )

    @run_in_background_thread('init-server-db')
    def initialize_server_db(self):
        """
        Load the lists of organisations and servers.
        """
        self.server_db.update()
        self.interface_transition_threadsafe('server_db_finished_loading')

    @property
    def network_state(self):
        """
        Get the current state of the network.
        """
        return self.network_state_machine.state

    @property
    def interface_state(self):
        """
        Get the current state of the interface.
        """
        return self.interface_state_machine.state

    def connect_state_transition_callbacks(self, obj):
        """
        Register all state transition callback methods decorated with
        `@transition_callback()` and `@transition_edge_callback()`
        of an object.
        """
        from .network import NetworkState
        self.network_state_machine.connect_object_callbacks(obj, NetworkState)
        from .interface.state import InterfaceState
        self.interface_state_machine.connect_object_callbacks(
            obj, InterfaceState)

    def network_transition(self, transition, *args, **kwargs):
        """
        Perform a transition on the network state.
        """
        logger.info(
            f'network transitioning: '
            f'{self.network_state} -> {transition}')
        try:
            self.network_state_machine.transition(
                transition, self, *args, **kwargs)
        except InvalidStateTransition:
            logger.error(
                f'invalid network state transition: '
                f'{self.network_state} -> {transition}')
        else:
            logger.info(
                f'network transitioned: '
                f'{transition} -> {self.network_state}')

    def interface_transition(self, transition, *args, **kwargs):
        """
        Perform a transition on the interface state.
        """
        logger.info(
            f'interface transitioning: '
            f'{self.interface_state} -> {transition}')
        try:
            self.interface_state_machine.transition(
                transition, self, *args, **kwargs)
        except InvalidStateTransition:
            logger.error(
                f'invalid interface state transition: '
                f'{self.interface_state} -> {transition}')
        else:
            logger.info(
                f'interface transitioned: '
                f'{transition} -> {self.interface_state}')

    def network_transition_threadsafe(self, transition, *args, **kwargs):
        """
        Threadsafe version of `network_transition()`.
        """
        transit = self.make_func_threadsafe(self.network_transition)
        transit(transition, *args, **kwargs)

    def interface_transition_threadsafe(self, transition, *args, **kwargs):
        """
        Threadsafe version of `network_transition()`.
        """
        transit = self.make_func_threadsafe(self.interface_transition)
        transit(transition, *args, **kwargs)



from typing import Optional, Callable
import threading
from functools import lru_cache, partial, wraps
from logging import getLogger
from os import path
from sys import prefix

logger = getLogger(__file__)


def get_logger(name_space: str):
    return getLogger(name_space)


@lru_cache(maxsize=1)
def get_prefix() -> str:
    """
    Returns the Python prefix where eduVPN is installed

    returns:
        path to Python installation prefix
    """
    target = 'share/eduvpn/builder/mainwindow.ui'
    local = path.dirname(path.dirname(path.abspath(__file__)))
    options = [local, path.expanduser('~/.local'), '/usr/local', prefix]
    for option in options:
        logger.debug(f"looking for '{target}' in '{option}'")
        if path.isfile(path.join(option, target)):
            return option
    raise Exception("Can't find eduVPN installation")


def custom_server_oauth_url(address):
    if not address.startswith(('http://', 'https://')):
        address = f'https://{address}'
    if not address.endswith('/'):
        address += '/'
    return address


def thread_helper(func: Callable, *, name: Optional[str] = None) -> threading.Thread:
    """
    Runs a function in a thread

    args:
        func (lambda): a function to run in the background
    """
    thread = threading.Thread(target=func, name=name)
    thread.daemon = True
    thread.start()
    return thread


def run_in_background_thread(name: Optional[str] = None):
    """
    Decorator for functions that must always run
    in a background thread.
    """
    def decorator(func):

        @wraps(func)
        def background_func(*args, **kwargs):
            thread_helper(partial(func, *args, **kwargs), name=name)

        return background_func

    return decorator


def run_in_main_gtk_thread(func):
    """
    Decorator for functions that must always run
    in the main GTK thread.
    """
    from gi.repository import GLib

    @wraps(func)
    def main_gtk_thread_func(*args, **kwargs):
        GLib.idle_add(partial(func, *args, **kwargs))

    return main_gtk_thread_func


def run_periodically(func: Callable[[], None],
                     interval: float,
                     name: Optional[str] = None,
                     ) -> Callable[[], None]:
    """
    Run a funtion periodically in a background thread.

    The given function is called every `interval` seconds,
    until either it returns False
    or until the returned cancel callback is called.
    """
    if name is None:
        name = 'run-periodically'
    event = threading.Event()

    @run_in_background_thread(name)
    def run_periodic_thread():
        while 1:
            if event.wait(interval):
                return
            elif func() is False:
                return

    run_periodic_thread()
    return event.set



from typing import Union, Optional, List
from requests_oauthlib import OAuth2Session
from ..state_machine import BaseState
from ..oauth2 import OAuthWebServer
from ..crypto import Validity
from ..app import Application
from ..server import (
    AnyServer, PredefinedServer, ConfiguredServer,
    SecureInternetServer, OrganisationServer, CustomServer, Profile)
from .error import translate_error
from . import event
from . import transition


class InterfaceState(BaseState):
    """
    Base class for all interface states.
    """

    def server_db_finished_loading(self, app: Application) -> 'InterfaceState':
        # Loading the server db doesn't normally change the interface state.
        return self

    def toggle_settings(self, app: Application) -> 'InterfaceState':
        # Toggling the settings page normally shows the settings page.
        return ConfigureSettings(self)

    def encountered_exception(self,
                              app: Application,
                              message: Union[str, Exception],
                              next_state: Optional['InterfaceState'] = None,
                              ) -> 'InterfaceState':
        return ErrorState(message, next_state)


class InitialInterfaceState(InterfaceState):
    """
    The state of the interface when the app starts.

    This is a transient state until
    the actual first state has been determined.
    """

    def found_active_connection(self,
                                app: Application,
                                server: ConfiguredServer,
                                validity: Optional[Validity],
                                ) -> InterfaceState:
        """
        An connection is already active, show its details.
        """
        return ConnectionStatus(server, validity)

    def no_active_connection_found(self, app: Application) -> InterfaceState:
        """
        No connection is currently active, go to the main state.

        The concrete main state depends on
        whether any servers have been configured previously.
        """
        return transition.go_to_main_state(app)


class MainState(InterfaceState):
    """
    The home state of the app,
    if any servers have been configured.

    Present the list of configured servers to start a connection.
    """

    def __init__(self, servers: List[ConfiguredServer]):
        self.servers = servers

    def configure_new_server(self, app: Application) -> InterfaceState:
        "Configure a new server."
        return ConfigurePredefinedServer()

    def connect_to_server(self,
                          app: Application,
                          server: ConfiguredServer) -> InterfaceState:
        "Connect to an already configured server."
        return transition.connect_to_server(app, server)


class PendingConfigurePredefinedServer(InterfaceState):
    """
    When the user searches for a server, but the list
    hasn't been downloaded yet, this state is temporarily
    set until the download is finished.
    """

    def __init__(self, search_query: str = ''):
        self.search_query = search_query

    def enter_search_query(self,
                           app: Application,
                           search_query: str,
                           ) -> InterfaceState:
        return PendingConfigurePredefinedServer(search_query)

    def enter_custom_address(self,
                             app: Application,
                             address: str,
                             ) -> InterfaceState:
        return transition.enter_custom_address(app, address)

    def connect_to_server(self,
                          app: Application,
                          server: ConfiguredServer) -> InterfaceState:
        return transition.connect_to_server(app, server)

    def server_db_finished_loading(self, app: Application) -> InterfaceState:
        """
        The list of predefined servers has been loaded,
        and the can how be shown to the user.
        """
        return transition.enter_search_query(app, self.search_query)


class ConfigurePredefinedServer(InterfaceState):
    """
    Select a server from a list of known servers to configure.
    """

    def __init__(self,
                 search_query: str = '',
                 results: Optional[List[PredefinedServer]] = None):
        self.search_query = search_query
        self.results = results

    def __repr__(self):
        if self.results is None:
            results = None
        else:
            # Don't include the long list of results.
            results = len(self.results)
        return (
            f"<ConfigurePredefinedServer"
            f" search_query={self.search_query!r}"
            f" results={results}>"
        )

    def enter_search_query(self, app: Application,
                           search_query: str,
                           ) -> InterfaceState:
        return transition.enter_search_query(app, search_query)

    def enter_custom_address(self, app: Application,
                             address: str,
                             ) -> InterfaceState:
        return transition.enter_custom_address(app, address)

    def connect_to_server(self,
                          app: Application,
                          server: PredefinedServer) -> InterfaceState:
        return transition.connect_to_server(app, server)


class ConfigureCustomServer(InterfaceState):
    """
    Select a custom server to configure
    by providing its address.
    """

    def __init__(self, address: str):
        self.address = address

    def enter_search_query(self, app: Application,
                           search_query: str,
                           ) -> InterfaceState:
        return transition.enter_search_query(app, search_query)

    def enter_custom_address(self, app: Application,
                             address: str,
                             ) -> InterfaceState:
        return transition.enter_custom_address(app, address)

    def connect_to_server(self,
                          app: Application,
                          server: CustomServer) -> InterfaceState:
        return transition.connect_to_server(app, server)


configure_server_states = (
    # This is a tuple so it can be used with `isinstance()`.
    PendingConfigurePredefinedServer,
    ConfigurePredefinedServer,
    ConfigureCustomServer,
)


class OAuthSetupPending(InterfaceState):
    """
    Wait for the local OAuth webserver to start.
    """

    def __init__(self, server: AnyServer):
        self.server = server

    def ready_for_oauth_setup(self,
                              app: Application,
                              oauth_web_server: OAuthWebServer) -> InterfaceState:
        """
        Cancel the OAuth setup process
        and take the user back to the main page.
        """
        return OAuthSetup(self.server, oauth_web_server)


class OAuthSetup(InterfaceState):
    """
    Allow the user to log into the VPN server using their browser.
    """

    def __init__(self, server: AnyServer, oauth_web_server: OAuthWebServer):
        self.server = server
        self.oauth_web_server = oauth_web_server

    def oauth_setup_cancel(self, app: Application) -> InterfaceState:
        """
        Cancel the OAuth setup process
        and take the user back to the main page.
        """
        self.oauth_web_server.stop()
        return transition.go_to_main_state(app)

    def oauth_setup_success(self,
                            app: Application,
                            oauth_session: OAuth2Session) -> InterfaceState:
        """
        The user has successfully completed the oauth setup by logging in.
        """
        event.on_refresh_oauth_token(app, self.server, oauth_session)
        return OAuthRefreshToken(app, self.server, oauth_session)


class OAuthRefreshToken(InterfaceState):
    """
    The OAuth token is being refreshed.
    """

    def __init__(self,
                 app: Application,
                 server: AnyServer,
                 oauth_session: OAuth2Session):
        self.app = app
        self.server = server
        self.oauth_session = oauth_session

    def oauth_refresh_success(self, app: Application) -> InterfaceState:
        event.on_start_connection(app, self.server, self.oauth_session)
        return LoadingServerInformation()

    def oauth_refresh_failed(self, app: Application) -> InterfaceState:
        """
        Refreshing the OAuth token failed,
        so the OAuth setup needs to be redone.
        """
        event.on_setup_oauth(app, self.server)
        return OAuthSetupPending(self.server)


class LoadingServerInformation(InterfaceState):
    """
    Wait for server information to be requested.
    """

    def choose_secure_internet_location(self,
                                        app: Application,
                                        server: OrganisationServer,
                                        oauth_session: OAuth2Session,
                                        locations: List[SecureInternetServer]):
        if len(locations) == 1:
            # Skip location choice if there's only a single option.
            event.on_start_connection(app, server, oauth_session, locations[0])
            return LoadingServerInformation()
        else:
            return ChooseSecureInternetLocation(server, oauth_session, locations)

    def choose_profile(self,
                       app: Application,
                       server: AnyServer,
                       oauth_session: OAuth2Session,
                       profiles: List[Profile]) -> InterfaceState:
        if len(profiles) == 1:
            # Skip profile choice if there's only a single option.
            event.on_chosen_profile(app, server, oauth_session, profiles[0])
            return ConfiguringConnection(server)
        else:
            return ChooseProfile(server, oauth_session, profiles)


class ChooseSecureInternetLocation(InterfaceState):
    """
    Allow the user to choose a secure internet location.
    """

    def __init__(self,
                 server: OrganisationServer,
                 oauth_session: OAuth2Session,
                 locations: List[SecureInternetServer]):
        self.server = server
        self.oauth_session = oauth_session
        self.locations = locations

    def select_secure_internet_location(self, app, location) -> InterfaceState:
        event.on_start_connection(app, self.server, self.oauth_session, location)
        return LoadingServerInformation()


class ChooseProfile(InterfaceState):
    """
    Allow the user to choose a profile.
    """

    def __init__(self,
                 server: AnyServer,
                 oauth_session: OAuth2Session,
                 profiles: List[Profile]):
        self.server = server
        self.oauth_session = oauth_session
        self.profiles = profiles

    def select_profile(self,
                       app: Application,
                       profile: Profile) -> InterfaceState:
        event.on_chosen_profile(app, self.server, self.oauth_session, profile)
        return ConfiguringConnection(self.server)


class ConfiguringConnection(InterfaceState):
    """
    Waiting to get the connection data and
    save the configuration to the network manager.
    """

    def __init__(self, server: AnyServer):
        self.server = server

    def finished_configuring_connection(self,
                                        app: Application,
                                        validity: Optional[Validity],
                                        ) -> InterfaceState:
        return ConnectionStatus(self.server, validity)


class ConnectionStatus(InterfaceState):
    """
    Show info on the active connection status.
    """

    def __init__(self,
                 server: AnyServer,
                 validity: Optional[Validity]):
        self.server = server
        self.validity = validity

    def go_back(self, app: Application) -> InterfaceState:
        return transition.go_to_main_state(app)

    def activate_connection(self, app: Application) -> InterfaceState:
        app.network_transition('reconnect')
        return self

    def deactivate_connection(self, app: Application) -> InterfaceState:
        app.network_transition('disconnect')
        return self

    def renew_certificate(self, app: Application) -> InterfaceState:
        return transition.connect_to_server(app, self.server, renew=True)


class ConfigureSettings(InterfaceState):
    """
    Allow the user to configure the application settings.
    """

    def __init__(self, previous_state):
        self.previous_state = previous_state

    def toggle_settings(self, app: Application) -> InterfaceState:
        return self.previous_state


class ErrorState(InterfaceState):
    """
    An error has occured.
    """

    def __init__(self, message: Union[str, Exception], next_state: Optional[InterfaceState] = None):
        if isinstance(message, Exception):
            message = translate_error(message)
        self.message = message
        self.next_state = next_state

    def acknowledge_error(self, app: Application) -> InterfaceState:
        if self.next_state is None:
            return transition.go_to_main_state(app)
        else:
            return self.next_state



from unittest import mock, TestCase
from eduvpn.oauth2 import get_oauth, one_request, get_open_port, stringify_image, OAuth2Session


class TestOauth(TestCase):

    @mock.patch.object(OAuth2Session, 'fetch_token')
    @mock.patch.object(OAuth2Session, 'authorization_url')
    @mock.patch('eduvpn.oauth2.one_request')
    @mock.patch('eduvpn.oauth2.webbrowser.open')
    def test_get_oauth(
            self,
            webbrowser_open: mock.MagicMock,
            one_request: mock.MagicMock,
            authorization_url: mock.MagicMock,
            fetch_token: mock.MagicMock,
    ):
        authorization_url.return_value = ["authorization_url", "state"]
        one_request.return_value = {'code': ["test"], 'state': ['state']}
        fetch_token.return_value = "token"

        get_oauth(authorization_endpoint="https://test", token_endpoint="https://test")

    def test_one_request(self):
        with self.assertRaises(Exception):
            one_request(port=9999, lets_connect=True, timeout=1)

        # one_request(port=9999, lets_connect=False)

    def test_get_open_port(self):
        get_open_port()

    def test_stringify_image(self):
        with self.assertRaises(FileNotFoundError):
            stringify_image("path")



from typing import (
    TypeVar, Generic, Any, Union, Optional, Callable,
    Type, Iterable, Tuple, Dict, Set)
import enum


class TransitionEdge(enum.Enum):
    """
    The edge of a state lifetime.

    The edge is `enter` when the state starts,
    and `exit` when it ends.
    """

    enter = enum.auto()
    exit = enum.auto()


ENTER = TransitionEdge.enter
EXIT = TransitionEdge.exit


# typing aliases
State = TypeVar('State')
StateType = Type[State]
Callback = Callable[[State, State], None]
StateTargets = Union[StateType, Iterable[StateType]]
CallbackRegistry = Dict[
    Optional[Tuple[StateType, TransitionEdge]],
    Set[Callback]]


TRANSITION_CALLBACK_MARKER = '__transition_callback_for_state'


def setattr_list_item(obj, attr, item):
    try:
        list_attr = getattr(obj, attr)
    except AttributeError:
        list_attr = []
        setattr(obj, attr, list_attr)
    list_attr.append(item)


def transition_callback(state_targets: StateTargets):
    """
    Decorator factory to mark a method as a
    transition callback for all transitions.

    Note the argument is the base class for states of the state machine
    to register transition events of.
    Without this, there would be no way to know which
    state machine this callback targets.
    """
    if not isinstance(state_targets, tuple):
        # Normalise argument to tuple.
        state_targets = (state_targets, )  # type: ignore

    def decorator(func: Callback):
        for state_type in state_targets:
            setattr_list_item(func,
                              TRANSITION_CALLBACK_MARKER,
                              (None, state_type))
        return func

    return decorator


def transition_edge_callback(edge: TransitionEdge,
                             state_targets: StateTargets):
    """
    Decorator factory to mark a method as a transition callback
    for specific state transition edges.
    """
    if not isinstance(state_targets, tuple):
        # Normalise argument to tuple.
        state_targets = (state_targets, )  # type: ignore

    def decorator(func: Callback):
        for state_type in state_targets:
            setattr_list_item(func,
                              TRANSITION_CALLBACK_MARKER,
                              (edge, state_type))
        return func

    return decorator


def _find_transition_callbacks(obj: Any, base_state_type: Type[State]):
    for attr in dir(obj):
        callback = getattr(obj, attr)
        try:
            registrations = getattr(callback, TRANSITION_CALLBACK_MARKER)
        except AttributeError:
            pass
        else:
            for edge, state_type in registrations:
                if issubclass(state_type, base_state_type):
                    yield callback, edge, state_type


class InvalidStateTransition(Exception):
    def __init__(self, name: str):
        self.name = name


class StateMachine(Generic[State]):
    """
    State machine wrapper that allows registering transition callbacks.
    """

    def __init__(self, initial_state: State):
        self._state = initial_state
        self._callbacks: CallbackRegistry = {}

    @property
    def state(self) -> State:
        """
        Obtain the current state.

        The state can be changed by calling `transition()`
        with the name of a transition of the current state.
        """
        # The state is behind a property to prevent setting it directly.
        return self._state

    def transition(self, transition: str, *args, **kwargs):
        """
        Transition to a new state.

        This method is *not* thread-safe,
        all calls should be made from the same thread.
        """
        old_state = self._state
        try:
            transition_func = getattr(old_state, transition)
        except AttributeError as e:
            raise InvalidStateTransition(transition) from e
        new_state = transition_func(*args, **kwargs)
        self._call_edge_callbacks(EXIT, old_state, new_state)
        self._state = new_state
        self._call_generic_callbacks(old_state, new_state)
        self._call_edge_callbacks(ENTER, old_state, new_state)
        return new_state

    def register_generic_callback(self, callback: Callback):
        """
        Register a callback for all transitions.
        """
        self._callbacks.setdefault(None, set()).add(callback)

    def register_edge_callback(self,
                               state_type: Type[State],
                               edge: TransitionEdge,
                               callback: Callback):
        """
        Register a callback for specific transition edges.
        """
        self._callbacks.setdefault((state_type, edge), set()).add(callback)

    def connect_object_callbacks(self, obj, base_state_type: Type[State]):
        """
        Register all state transition callback methods decorated with
        `@transition_callback()` and `@transition_edge_callback()`
        of an object.

        Provide the base class of states for this state machine
        as the second argument to filter registrations for this
        state machine only.
        Only method registered to a subclass of this base class
        will be connected.
        """
        iterator = _find_transition_callbacks(obj, base_state_type)
        for callback, edge, state_type in iterator:
            if edge is None:
                # This callback targets all events.
                self.register_generic_callback(callback)
            else:
                # This callback targets a specific state edge.
                self.register_edge_callback(state_type, edge, callback)

    def _call_generic_callbacks(self, old_state: State, new_state: State):
        for callback in self._callbacks.get(None, []):
            callback(old_state, new_state)

    def _call_edge_callbacks(self,
                             edge: TransitionEdge,
                             old_state: State,
                             new_state: State):
        state = new_state if edge is ENTER else old_state
        for callback in self._callbacks.get((state.__class__, edge), []):
            callback(old_state, new_state)


class BaseState:
    """
    Base class for all state machine states.
    """

    def __repr__(self):
        fields = ','.join(f' {k}={v!r}' for k, v in self.__dict__.items())
        return f'<{self.__class__.__name__}{fields}>'

    def has_transition(self, name: str) -> bool:
        """
        Return True if this state defines the transition function.
        """
        return hasattr(self, name)

    def copy(self, **fields):
        """
        Return a copy of this state, with some fields altered.
        """
        return self.__class__(**{**self.__dict__, **fields})



from unittest import TestCase, skipIf

from eduvpn.nm import (nm_available, add_connection,
                       import_ovpn, get_client, get_mainloop, activate_connection,
                       get_cert_key, connection_status, deactivate_connection, save_connection, update_connection)
from eduvpn.storage import get_uuid
from tests.mock_config import mock_config, mock_key, mock_cert


@skipIf(not nm_available(), "Network manager not available")
class TestNm(TestCase):
    def test_nm_available(self):
        nm_available()

    def test_import_ovpn(self):
        simple_connection = import_ovpn(mock_config, mock_key, mock_cert)

    def test_get_mainloop(self):
        main_loop = get_mainloop()

    def test_get_add_connection(self):
        client = get_client()
        simple_connection = import_ovpn(mock_config, mock_key, mock_cert)
        add_connection(client, simple_connection)

    def test_get_uuid(self):
        uuid = get_uuid()

    def test_activate_connection(self):
        client = get_client()
        uuid = get_uuid()
        # activate_connection(client, uuid)
        # get_cert_key(client, uuid)
        # connection_status(client, uuid)
        # deactivate_connection(client, uuid)
        # save_connection(client, mock_config, mock_key, mock_cert)



from logging import getLogger
from typing import List
from enum import Enum
from oauthlib.oauth2.rfc6749.tokens import OAuth2Token

logger = getLogger(__name__)


class VpnConnection:
    class ConnectionType(str, Enum):
        INSTITUTE = "INSTITUTE",
        SECURE = "SECURE",
        OTHER = "OTHER"

    def __init__(self):
        self.type: str = VpnConnection.ConnectionType.OTHER
        self.api_url: str = ""
        self.auth_url: str = ""
        self.token_endpoint: str = ""
        self.authorization_endpoint: str = ""
        self.server_name: str = ""
        self.server_image = None
        self.profile_id: str = ""
        self.support_contact: List[str] = []
        self.token: OAuth2Token

    def __str__(self):
        return str(self.__class__) + ": " + str(self.__dict__)



from argparse import Namespace
from unittest import TestCase, mock
from unittest.mock import patch, MagicMock
from tests.mock_config import mock_server, mock_org
from eduvpn.menu import menu, input_int, profile_choice, provider_choice, write_to_nm_choice
from eduvpn.menu import search, configure, interactive, match_term, fetch_servers_orgs


class TestMenu(TestCase):
    def test_menu(self):
        with mock.patch('builtins.input', lambda _: '0'):
            menu(institutes=[{'display_name': 'test', 'base_url': 'no url', 'support_contact': 'rutte@gov.nl'}], orgs=[], search_term="test")

    def test_input_int(self):
        with mock.patch('builtins.input', lambda _: '1'):
            input_int(max_=3)

    def test_profile_choice(self):
        profiles = [{'profile_id': 'internet'}]
        profile_choice(profiles=profiles)

    def test_provider_choice(self):
        base_uri = 'bla'
        institutes = [{'display_name': 'test', 'base_url': base_uri, 'support_contact': 'trump@whitehouse.gov'}]
        with mock.patch('builtins.input', lambda _: '0'):
            url, display_name, contact, secure_internet = provider_choice(institutes=institutes, orgs=[])
        self.assertEqual(secure_internet, False)
        self.assertEqual(base_uri, url)

    def test_write_to_nm_choice(self):
        with mock.patch('builtins.input', lambda _: '1'):
            write_to_nm_choice()

    @patch('eduvpn.menu.fetch_servers_orgs')
    @patch('eduvpn.actions.fetch_token')
    def test_configure(self, _: MagicMock, fetch_servers_orgs_: MagicMock):
        fetch_servers_orgs_.return_value = [mock_server], [mock_org]
        configure(Namespace(match='bogus'))
        configure(Namespace(match=''))

    def test_match_term(self):
        match_term(servers=[], orgs=[], search_term="search")

    @patch('eduvpn.menu.fetch_servers_orgs')
    def test_search(self, fetch_servers_orgs_: MagicMock):
        fetch_servers_orgs_.return_value = [mock_server], [mock_org]
        search(Namespace(match='bogus'))

    @patch('eduvpn.menu.list_servers')
    @patch('eduvpn.menu.list_organisations')
    def test_fetch_servers_orgs(self, list_organisations, list_servers):
        fetch_servers_orgs()

    def test_main_with_url(self):
        args = MagicMock()
        args.match = "https://test"
        interactive(args)



from logging import getLogger
from typing import Tuple

from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.errors import InvalidGrantError

from eduvpn.menu import profile_choice
from eduvpn.nm import activate_connection_with_mainloop, get_cert_key, get_client,\
    nm_available, connection_status, save_connection_with_mainloop, deactivate_connection_with_mainloop
from eduvpn.oauth2 import get_oauth
from eduvpn.remote import get_info, check_certificate, create_keypair, get_config, list_profiles
from eduvpn.settings import CLIENT_ID
from eduvpn.storage import get_storage, get_current_metadata, update_token, get_uuid, get_all_metadatas

_logger = getLogger(__file__)


def refresh():
    """
    Refreshes an active configuration. The token is refreshed if expired, and a new token is obtained if the token
    is invalid.
    """
    uuid, auth_url, metadata = get_storage(check=True)
    token, token_endpoint, auth_endpoint, api_url, display_name, support_contact, profile_id, con_type, country_id, _, _ = metadata
    oauth = OAuth2Session(client_id=CLIENT_ID, token=token, auto_refresh_url=token_endpoint)

    try:
        token = oauth.refresh_token(token_url=token_endpoint)
    except InvalidGrantError as e:
        _logger.warning(f"token invalid: {e}")
        oauth = get_oauth(token_endpoint, auth_endpoint)

    api_base_uri, token_endpoint, auth_endpoint = get_info(auth_url)
    client = get_client()
    try:
        cert, key = get_cert_key(client, uuid)
    except IOError:
        # probably the NM connection was deleted
        cert = None

    if not cert or not check_certificate(oauth, api_base_uri, cert):
        key, cert = create_keypair(oauth, api_base_uri)
        config = get_config(oauth, api_base_uri, profile_id)
        save_connection_with_mainloop(config, key, cert)

    update_token(token)


def fetch_token(auth_url: str) -> Tuple[str, OAuth2Session, str, str]:
    """
    Starts the full enrollment procedure.

    Once completed, will store api_url, auth_url and profile in storage. Returns config, private key and certifcate,
    which you should use to compose a valid VPN configuration and store in NM or in a ovpn file.
    """
    # make sure our URL ends with a /
    if auth_url[-1] != '/':
        auth_url += '/'

    _logger.info(f"starting procedure with auth_url {auth_url}")
    exists = get_current_metadata(auth_url)

    if exists:
        _logger.info("token exists, restoring")
        token, token_endpoint, auth_endpoint, api_url, display_name, support_contact, profile_id, con_type, country_id, _, _ = exists
        oauth = OAuth2Session(client_id=CLIENT_ID, token=token, auto_refresh_url=token_endpoint)
        api_url, token_endpoint, auth_endpoint = get_info(auth_url)
    else:
        _logger.info("fetching token")
        api_url, token_endpoint, auth_endpoint = get_info(auth_url)
        oauth = get_oauth(token_endpoint, auth_endpoint)

    try:
        oauth.refresh_token(token_url=token_endpoint)
    except InvalidGrantError as e:
        _logger.warning(f"token invalid: {e}")
        oauth = get_oauth(token_endpoint, auth_endpoint)

    return api_url, oauth, token_endpoint, auth_endpoint


def get_profile(oauth: OAuth2Session, api_url: str, interactive: bool = False):
    profiles = list_profiles(oauth, api_url)
    if interactive:
        return profile_choice(profiles)
    else:
        return profile_choice(profiles[:1])


def get_config_and_keycert(oauth: OAuth2Session, api_url: str, profile_id: str) -> Tuple[str, str, str]:
    config = get_config(oauth, api_url, profile_id)
    private_key, certificate = create_keypair(oauth, api_url)
    return config, private_key, certificate


def status():
    uuid, current_auth_url, metadata = get_storage(check=False)

    print("\n\n# Global configuration\n")
    print(f"uuid: {uuid}")
    print(f"auth_url: {current_auth_url}")

    if nm_available():
        active_uuid, status = connection_status(get_client())
        print(f"VPN connection active: {bool(active_uuid)}")
        print(f"VPN NM status is: {status}")
        print(f"Active VPN connection is EduVPN: {bool(uuid == active_uuid)}")

    print("\n\n## Previous connection properties\n")

    for auth_url, props in get_all_metadatas().items():
        if auth_url == current_auth_url:
            print(" ** CURRENT ACTIVE CONFIGURATION **")
        print(f"auth_url {auth_url}")
        for t in ['api_url', 'display_name', 'support_contact', 'profile_id',
                  'token_endpoint', 'authorization_endpoint', 'con_type', 'country_id']:
            print(f"{t}: {props[t]}")
        print(f"token.token_type: {props['token']['token_type']}")
        print(f"token.expires_in: {props['token']['expires_in']}")
        print(f"token.expires_at: {props['token']['expires_at']}")
        print("\n\n")


def activate():
    """
    Activates an existing configuration
    """
    refresh()
    activate_connection_with_mainloop(get_uuid())


def deactivate():
    """
    Deactivates an existing configuration
    """
    deactivate_connection_with_mainloop(get_uuid())



from typing import Optional
import logging
import enum
from functools import partial
from time import sleep
from gettext import gettext
from . import nm
from .crypto import Validity
from . import settings
from .state_machine import BaseState
from .app import Application
from .server import ConfiguredServer as Server
from .utils import run_in_background_thread

logger = logging.getLogger(__name__)


def translated_property(text):
    return property(lambda self: gettext(text))  # type: ignore


class StatusImage(enum.Enum):
    # The value is the image filename.
    DEFAULT = 'desktop-default.png'
    CONNECTING = 'desktop-connecting.png'
    CONNECTED = 'desktop-connected.png'
    NOT_CONNECTED = 'desktop-not-connected.png'

    @property
    def path(self):
        return settings.IMAGE_PREFIX + self.value


class NetworkState(BaseState):
    """
    Base class for all interface states.
    """

    status_label = translated_property("Connection state unknown")
    status_image = StatusImage.DEFAULT

    def start_new_connection(self,
                             app: Application,
                             server: Server,
                             ) -> 'NetworkState':
        return connect(app)

    def set_connecting(self, app: Application) -> 'NetworkState':
        return ConnectingState()

    def set_connected(self, app: Application) -> 'NetworkState':
        return ConnectedState()

    def set_disconnected(self, app: Application) -> 'NetworkState':
        return DisconnectedState()

    def set_unknown(self, app: Application) -> 'NetworkState':
        return enter_unknown_state(app)

    def set_error(self, app: Application, message: Optional[str] = None) -> 'NetworkState':
        return ConnectionErrorState(message)

    def set_certificate_expired(self, app: Application) -> 'NetworkState':
        return CertificateExpiredState()


class InitialNetworkState(NetworkState):
    """
    The state of the network when the app starts.

    This is a transient state until
    the actual network state is obtained.
    """

    def found_active_connection(self,
                                app: Application,
                                server: Server,
                                validity: Optional[Validity],
                                ) -> NetworkState:
        """
        An already active connection was found.
        """
        app.interface_transition('found_active_connection', server, validity)
        return ConnectedState()

    def found_previous_connection(self,
                                  app: Application,
                                  server: Server,
                                  ) -> NetworkState:
        """
        A previously established connection was found.

        This will be the default connection to start.
        """
        app.interface_transition('no_active_connection_found')
        return DisconnectedState()

    def no_previous_connection_found(self, app: Application) -> NetworkState:
        """
        No previously established connection was found.

        This is probably the first time the app runs.
        """
        app.interface_transition('no_active_connection_found')
        return UnconnectedState()


class UnconnectedState(NetworkState):
    """
    There is no current connection active,
    nor is there a configured connection available.

    As soon as a server is chosen,
    that becomes the default
    and this state is no longer reached.
    """

    status_label = translated_property("Disconnected")


def connect(app: Application) -> NetworkState:
    """
    Estabilish a connection to the server.
    """
    client = nm.get_client()
    nm.activate_connection(
        client,
        app.current_network_uuid,
        partial(on_any_update_callback, app),
    )
    return ConnectingState()


def disconnect(app: Application, *, update_state=True) -> NetworkState:
    """
    Break the connection to the server.
    """
    client = nm.get_client()
    callback = None
    if update_state:
        callback = partial(on_any_update_callback, app)
    nm.deactivate_connection(
        client,
        app.current_network_uuid,
        callback,
    )
    return DisconnectedState()


def reconnect(app: Application) -> NetworkState:
    """
    Disconnect and then connect again.
    """
    disconnect(app, update_state=False)
    return ReconnectingState()


def renew_certificate(app: Application) -> NetworkState:
    app.interface_transition('renew_certificate')
    return ConnectedState()


UNKNOWN_STATE_MAX_RETRIES = 5


def enter_unknown_state(app: Application) -> NetworkState:
    # Set the state temporarily to unknown but keep polling for updates,
    # since we don't always get notified by the update callback.
    @run_in_background_thread('poll-network-state')
    def determine_network_state_thread():
        counter = 0
        _, status = nm.connection_status(nm.get_client())
        while status is nm.NM.ActiveConnectionState.UNKNOWN or status is None:
            if counter > UNKNOWN_STATE_MAX_RETRIES:
                # After a number of retries, assume we've disconnected
                # so the user can try to connect again.
                status = nm.NM.ActiveConnectionState.DEACTIVATED
                logger.debug(
                    "network state has been unknown for too long,"
                    " fall back to disconnected state"
                )
                break
            sleep(1)
            counter += 1
            if not isinstance(app.network_state, UnknownState):
                return
            _, status = nm.connection_status(nm.get_client())
            logger.debug(f"polling network state: {status}")
        handle_active_connection_status(app, status)

    determine_network_state_thread()
    return UnknownState()


def on_status_update_callback(app: Application, status: nm.NM.VpnConnectionState):
    """
    Callback for whenever a connection status changes.
    """
    if status in [nm.NM.VpnConnectionState.CONNECT,
                  nm.NM.VpnConnectionState.IP_CONFIG_GET,
                  nm.NM.VpnConnectionState.PREPARE]:
        app.network_transition('set_connecting')
    elif status is nm.NM.VpnConnectionState.ACTIVATED:
        app.network_transition('set_connected')
    elif status is nm.NM.VpnConnectionState.DISCONNECTED:
        app.network_transition('set_disconnected')
    elif status is nm.NM.VpnConnectionState.FAILED:
        app.network_transition('set_error')
    elif status is nm.NM.VpnConnectionState.NEED_AUTH:
        app.network_transition('set_error')
    elif status is nm.NM.VpnConnectionState.UNKNOWN:
        app.network_transition('set_unknown')
    else:
        raise ValueError(status)


def on_any_update_callback(app: Application):
    """
    Callback for whenever a connection status might have changed.
    """
    _, status = nm.connection_status(nm.get_client())
    if status is None:
        app.network_transition('set_unknown')
    else:
        handle_active_connection_status(app, status)


def handle_active_connection_status(app: Application, status: nm.NM.ActiveConnectionState):
    if status is nm.NM.ActiveConnectionState.ACTIVATING:
        app.network_transition('set_connecting')
    elif status is nm.NM.ActiveConnectionState.ACTIVATED:
        app.network_transition('set_connected')
    elif status in [nm.NM.ActiveConnectionState.DEACTIVATED,
                    nm.NM.ActiveConnectionState.DEACTIVATING]:
        app.network_transition('set_disconnected')
    elif status is nm.NM.ActiveConnectionState.UNKNOWN:
        app.network_transition('set_unknown')
    else:
        raise ValueError(status)


class ConnectingState(NetworkState):
    """
    The network is currently trying to connect to a server.
    """

    status_label = translated_property("Preparing to connect")
    status_image = StatusImage.CONNECTING

    def start_new_connection(self,
                             app: Application,
                             server: Server,
                             ) -> NetworkState:
        return reconnect(app)

    def disconnect(self, app: Application) -> NetworkState:
        """
        Abort connecting.
        """
        return disconnect(app)


class ConnectedState(NetworkState):
    """
    The network is currently connected to a server.
    """

    status_label = translated_property("Connection active")
    status_image = StatusImage.CONNECTED

    def start_new_connection(self,
                             app: Application,
                             server: Server,
                             ) -> NetworkState:
        return reconnect(app)

    def disconnect(self, app: Application) -> NetworkState:
        return disconnect(app)

    def renew_certificate(self, app: Application) -> NetworkState:
        """
        Re-estabilish a connection to the server.
        """
        return renew_certificate(app)


class DisconnectedState(NetworkState):
    """
    The network is not currently connected to a server,
    but a configured connection exists.
    """

    status_label = translated_property("Disconnected")
    status_image = StatusImage.NOT_CONNECTED

    def reconnect(self, app: Application) -> NetworkState:
        return connect(app)


class ReconnectingState(NetworkState):
    """
    The network is currently disconnecting
    and will connect again when ready.
    """

    status_label = translated_property("Preparing to connect")
    status_image = StatusImage.CONNECTING

    def set_unknown(self, app: Application) -> NetworkState:
        return connect(app)

    def set_disconnected(self, app: Application) -> NetworkState:
        return connect(app)


class CertificateExpiredState(NetworkState):
    """
    The network could not connect because the certifcate has expired.
    """

    status_label = translated_property("Certificate expired")
    status_image = StatusImage.NOT_CONNECTED

    def renew_certificate(self, app: Application) -> NetworkState:
        """
        Re-estabilish a connection to the server.
        """
        return renew_certificate(app)


class ConnectionErrorState(NetworkState):
    """
    The network could not connect because an error occured.
    """

    status_label = translated_property("Connection failed")
    status_image = StatusImage.NOT_CONNECTED

    def __init__(self, error: Optional[str]):
        self.error = error

    def reconnect(self, app: Application) -> NetworkState:
        return connect(app)


class UnknownState(NetworkState):
    """
    The network state could not be determined.
    """

    def reconnect(self, app: Application) -> NetworkState:
        return connect(app)



import logging
import webbrowser
from typing import Optional, Callable, Dict, List
from http.server import BaseHTTPRequestHandler, HTTPServer
import socket
from urllib.parse import urlparse, parse_qs
import requests
from requests_oauthlib import OAuth2Session
from eduvpn.settings import get_brand
from eduvpn.settings import CLIENT_ID, SCOPE, CODE_CHALLENGE_METHOD
from eduvpn.utils import run_in_background_thread

from eduvpn.crypto import gen_code_verifier, gen_code_challenge

logger = logging.getLogger(__name__)

landing_page = """
<!doctype html>
<html lang=en>
<head>
<meta charset=utf-8>
<title>{brand} - bye</title>
<style>
.center {{
    font-family: arial;
    font-size: 30px;
    position: absolute;
    text-align: center;
    width: 800px;
    height: 50px;
    top: 50%;
    left: 50%;
    margin-left: -400px; /* margin is -0.5 * dimension */
    margin-top: -25px;
}}
</style>
</head>
<body>

<div class="center">
<img src="data:image/png;base64,{logo}" width="150">
<p>You can now close this window.</p>
</div>
</body>
</html>
"""


def stringify_image(logo: str) -> str:
    import base64
    return base64.b64encode(open(logo, 'rb').read()).decode('ascii')


def get_open_port() -> int:
    """
    Find an unused local port.

    returns:
        int: an unused port number
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def one_request(port: int, lets_connect: bool, timeout: Optional[int] = None) -> Optional[Dict[str, List[str]]]:
    """
    Listen for one http request on port, then close and return request query

    args:
        port (int): the port to listen for the request
    returns:
        str: the request
    """
    logger.info(f"listening for a request on port {port}...")

    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

            logo, name = get_brand(lets_connect)
            logo = stringify_image(logo)
            content = landing_page.format(logo=logo, brand=name).encode('utf-8')
            self.wfile.write(content)
            self.server.path = self.path

    httpd = HTTPServer(('', port), RequestHandler)
    if timeout:
        httpd.socket.settimeout(timeout)
    httpd.handle_request()
    httpd.server_close()

    if not hasattr(httpd, "path"):
        raise Exception("Invalid response received")

    path = httpd.path  # type: ignore
    if path == '/cancel':
        return None
    else:
        parsed = urlparse(path)
        logger.info(f"received a request {path}")  # type: ignore
        return parse_qs(parsed.query)


def get_oauth(token_endpoint: str, authorization_endpoint: str) -> Optional[OAuth2Session]:
    port = get_open_port()
    return get_oauth_at_port(port, token_endpoint, authorization_endpoint)


def get_oauth_at_port(port: int, token_endpoint: str, authorization_endpoint: str) -> Optional[OAuth2Session]:
    redirect_uri = f'http://127.0.0.1:{port}/callback'
    oauth = OAuth2Session(CLIENT_ID, redirect_uri=redirect_uri, auto_refresh_url=token_endpoint, scope=SCOPE)

    code_verifier = gen_code_verifier()
    code_challenge = gen_code_challenge(code_verifier)
    authorization_url, state = oauth.authorization_url(url=authorization_endpoint,
                                                       code_challenge_method=CODE_CHALLENGE_METHOD,
                                                       code_challenge=code_challenge)

    logger.info(f"opening browser with {authorization_url}")
    webbrowser.open(authorization_url)
    response = one_request(port, lets_connect=False)
    if response is None:
        return None
    code = response['code'][0]
    assert (state == response['state'][0])
    oauth.fetch_token(token_url=token_endpoint, code=code,
                      code_verifier=code_verifier,
                      client_id=oauth.client_id, include_client_id=True)
    return oauth


def send_cancel_request(port):
    requests.get(f'http://127.0.0.1:{port}/cancel')


class OAuthWebServer:
    def __init__(self, port: int):
        self.port = port

    @classmethod
    def start(cls,
              token_endpoint: str,
              authorization_endpoint: str,
              callback: Callable[[Optional[OAuth2Session]], None]) -> 'OAuthWebServer':
        port = get_open_port()

        @run_in_background_thread('oauth-http-server')
        def run_webserver():
            token = get_oauth_at_port(port, token_endpoint, authorization_endpoint)
            callback(token)

        run_webserver()
        return cls(port)

    @run_in_background_thread('oauth-http-server-stop')
    def stop(self):
        send_cancel_request(self.port)



from typing import Tuple
from pathlib import Path
from eduvpn.utils import get_prefix

prefix = get_prefix()

CONFIG_PREFIX = Path("~/.config/eduvpn/").expanduser().resolve()
CONFIG_JSON_PREFIX = "2.0_"

DISCO_URI = 'https://disco.eduvpn.org/v2/'
ORGANISATION_URI = DISCO_URI + "organization_list.json"
SERVER_URI = DISCO_URI + "server_list.json"
HELP_URL = 'https://www.eduvpn.org'

CLIENT_ID = "org.eduvpn.app.linux"
SCOPE = ["config"]
CODE_CHALLENGE_METHOD = "S256"
LANGUAGE = 'nl'
COUNTRY = "nl-NL"

COUNTRY_MAP = Path(prefix + "/share/eduvpn/country_codes.json")
FLAG_PREFIX = prefix + "/share/eduvpn/images/flags/png/"
IMAGE_PREFIX = prefix + "/share/eduvpn/images/"
LC_IMAGE_PREFIX = prefix + "/share/letsconnect/images/"


# format: base64(<signature_algorithm> || <key_id> || <public_key>)
VERIFY_KEYS = [
    "RWRtBSX1alxyGX+Xn3LuZnWUT0w//B6EmTJvgaAxBMYzlQeI+jdrO6KF",
    "RWQ68Y5/b8DED0TJ41B1LE7yAvkmavZWjDwCBUuC+Z2pP9HaSawzpEDA",
    "RWQKqtqvd0R7rUDp0rWzbtYPA3towPWcLDCl7eY9pBMMI/ohCmrS0WiM"
]
EDUVPN_ICON = prefix + "/share/icons/hicolor/128x128/apps/eduvpn-client.png"
EDUVPN_NAME = "eduVPN"
LETS_CONNECT_LOGO = LC_IMAGE_PREFIX + "letsconnect.png"
SERVER_ILLUSTRATION = LC_IMAGE_PREFIX + "server-illustration.png"
LETS_CONNECT_ICON = prefix + "/share/icons/hicolor/128x128/apps/letsconnect-client.png"
LETS_CONNECT_NAME = "Let's Connect!"


def get_brand(lets_connect: bool) -> Tuple[str, str]:
    """
    args:
        lets_connect: Set true if we are let's connect, otherwise eduVPN
    returns:
        logo, name
    """
    if lets_connect:
        return LETS_CONNECT_ICON, LETS_CONNECT_NAME
    else:
        return EDUVPN_ICON, EDUVPN_NAME



# eduvpngui - The GNU/Linux eduVPN GUI client
#
# Copyright: 2017-2020, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Any, Optional
import os
import webbrowser
import logging
from datetime import datetime
from gettext import gettext as _, ngettext

import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
gi.require_version('NM', '1.0')  # noqa: E402
from gi.repository import Gtk, GObject, GdkPixbuf

from ..settings import HELP_URL
from ..interface import state as interface_state
from .. import network as network_state
from ..server import CustomServer
from ..app import Application
from ..state_machine import (
    ENTER, EXIT, transition_callback, transition_edge_callback)
from ..crypto import Validity
from ..utils import get_prefix, run_in_main_gtk_thread, run_periodically
from ..i18n import init as i18n_init
from . import search
from .utils import show_ui_component, link_markup

logger = logging.getLogger(__name__)

builder_files = ['mainwindow.ui']

variable_objects = [
    'backButton',
    # 'backButtonEventBox',
    'findYourInstitutePage',
    'instituteTreeView',
    'secureInternetTreeView',
    'otherServersTreeView',
    'findYourInstituteSpacer',
    'findYourInstituteImage',
    'findYourInstituteLabel',
    'addOtherServerRow',
    'addOtherServerButton',
    'findYourInstituteScrolledWindow',
    'instituteAccessHeader',
    'secureInternetHeader',
    'otherServersHeader',
    'findYourInstituteSearch',
    'chooseProfilePage',
    'profileTreeView',
    'chooseLocationPage',
    'locationTreeView',
    'openBrowserPage',
    'connectionPage',
    'serverLabel',
    'serverImage',
    'supportLabel',
    'connectionStatusImage',
    'connectionStatusLabel',
    'connectionSubStatus',
    'profilesSubPage',
    'currentConnectionSubPage',
    'connectionSubPage',
    'connectionInfoTopRow',
    'connectionInfoGrid',
    'durationValueLabel',
    'downloadedValueLabel',
    'uploadedValueLabel',
    'ipv4ValueLabel',
    'ipv6ValueLabel',
    'connectionInfoBottomRow',
    'connectionSwitch',
    'settingsPage',
    'messagePage',
    'messageLabel',
    'messageText',
    'messageButton',
]


UPDATE_EXIPRY_INTERVAL = 1.  # seconds

RENEWAL_ALLOW_FRACTION = .8


def get_validity_text(validity: Optional[Validity]) -> str:
    if validity is None:
        return _("Valid for <b>unknown</b>")
    expiry = validity.end
    now = datetime.utcnow()
    if expiry <= now:
        return _("This session has expired")
    delta = expiry - now
    days = delta.days
    hours = delta.seconds // 3600
    if days == 0:
        if hours == 0:
            minutes = delta.seconds // 60
            return ngettext("Valid for <b>{0} minute</b>",
                            "Valid for <b>{0} minutes</b>", minutes).format(minutes)
        else:
            return ngettext("Valid for <b>{0} hour</b>",
                            "Valid for <b>{0} hours</b>", hours).format(hours)
    else:
        dstr = ngettext("Valid for <b>{0} day</b>",
                        "Valid for <b>{0} days</b>", days).format(days)
        hstr = ngettext(" and <b>{0} hour</b>",
                        " and <b>{0} hours</b>", hours).format(hours)
        return dstr + hstr


def allow_certificate_renewal(validity: Optional[Validity]):
    if validity is None:
        return True
    return datetime.utcnow() >= validity.fraction(RENEWAL_ALLOW_FRACTION)


class EduVpnGui:
    def __init__(self, lets_connect: bool):
        """
        Initialize all data structures needed in the GUI
        """
        self.lets_connect = lets_connect

        self.builder: Any = Gtk.Builder()

        prefix = get_prefix()
        try:
            self.builder.set_translation_domain(i18n_init(lets_connect, prefix))
            logger.info(u"i18n successfully initialized")
        except Exception as e:
            logger.error(f"i18n initialization failed: {e}")

        for b in builder_files:
            p = os.path.join(prefix, 'share/eduvpn/builder', b)
            if not os.access(p, os.R_OK):
                logger.error(f"Can't find builder file {p}!")
                raise Exception
            self.builder.add_from_file(p)

        handlers = {
            "delete_window": Gtk.main_quit,
            "on_settings_button_released": self.on_settings_button_released,
            "on_help_button_released": self.on_help_button_released,
            "on_back_button_released": self.on_back_button_released,
            "on_search_changed": self.on_search_changed,
            "on_activate_changed": self.on_activate_changed,
            "on_add_other_server_button_clicked":
                self.on_add_other_server_button_clicked,
            "on_cancel_browser_button_clicked":
                self.on_cancel_oauth_setup_button_clicked,
            "on_connection_switch_state_set":
                self.on_connection_switch_state_set,
            "on_renew_session_clicked":
                self.on_renew_session_clicked,
        }
        self.builder.connect_signals(handlers)

        # Hide all objects that are visible on only some pages.
        for name in variable_objects:
            self.show_component(name, False)

        self.app = Application(run_in_main_gtk_thread)

        # TODO implement settings page (issue #334)
        self.show_component('settingsButton', False)

    def run(self):
        logger.info("starting ui")
        self.show_component('applicationWindow', True)
        self.app.connect_state_transition_callbacks(self)
        self.app.initialize()

    # ui functions

    def show_component(self, component_id: str, show: bool):
        show_ui_component(self.builder, component_id, show)

    def set_text(self, component_id: str, text: str):
        component = self.builder.get_object(component_id)
        component.set_text(text)

    def set_markup_text(self, component_id: str, text: str):
        component = self.builder.get_object(component_id)
        component.set_markup(text)

    def show_back_button(self, show: bool):
        self.show_component('backButton', show)
        self.show_component('backButtonEventBox', show)

    def set_search_text(self, text: str):
        self.set_text('findYourInstituteSearch', text)

    def show_message_page(self, title: str, message: str):
        self.show_component('loadingPage', True)
        self.show_component('loadingSpacer', True)
        self.show_component('loadingTitle', True)
        self.show_component('loadingMessage', True)
        self.show_component('loadingSpinner', True)
        self.set_text('loadingTitle', title)
        self.set_text('loadingMessage', message)

    def hide_message_page(self):
        self.show_component('loadingPage', False)

    # network state transition callbacks

    @transition_callback(network_state.NetworkState)
    def default_network_transition_callback(self, old_state, new_state):
        if isinstance(self.app.interface_state, interface_state.ConnectionStatus):
            self.update_connection_status()

    def update_connection_server(self):
        server = self.app.interface_state.server
        self.show_component('serverLabel', True)
        self.set_text('serverLabel', str(server))

        server_image_component = self.builder.get_object('serverImage')
        server_image_path = getattr(server, 'image_path', None)
        if server_image_path:
            server_image_component.set_from_file(server_image_path)
            server_image_component.show()
        else:
            server_image_component.hide()

        if getattr(server, 'support_contact', []):
            support_text = _("Support:") + "\n" + "\n".join(map(link_markup, server.support_contact))
            self.set_markup_text('supportLabel', support_text)
            self.show_component('supportLabel', True)
        else:
            self.show_component('supportLabel', False)

    def update_connection_validity(self):
        expiry_text = get_validity_text(self.app.interface_state.validity)
        self.set_markup_text('connectionSubStatus', expiry_text)

        if (hasattr(self.app.network_state, 'renew_certificate') and (
                isinstance(self.app.network_state, network_state.CertificateExpiredState) or (
                allow_certificate_renewal(self.app.interface_state.validity)))):
            self.show_component('currentConnectionSubPage', True)
            self.show_component('renewSessionButton', True)
        else:
            self.show_component('renewSessionButton', False)

    def update_connection_status(self):
        self.set_text('connectionStatusLabel', self.app.network_state.status_label)
        self.builder.get_object('connectionStatusImage').set_from_file(self.app.network_state.status_image.path)

        self.update_connection_validity()

        assert not (hasattr(self.app.network_state, 'reconnect') and hasattr(self.app.network_state, 'disconnect'))
        connection_switch = self.builder.get_object('connectionSwitch')
        if hasattr(self.app.network_state, 'reconnect'):
            self.show_component('currentConnectionSubPage', True)
            connection_switch.show()
            connection_switch.set_state(False)
        elif hasattr(self.app.network_state, 'disconnect'):
            self.show_component('currentConnectionSubPage', True)
            connection_switch.show()
            connection_switch.set_state(True)
        else:
            self.show_component('currentConnectionSubPage', False)
            connection_switch.hide()

    # interface state transition callbacks

    @transition_callback(interface_state.InterfaceState)
    def default_interface_transition_callback(self, old_state, new_state):
        # Only show the 'go back' button if
        # the corresponding transition is available.
        self.show_back_button(new_state.has_transition('go_back'))

    @transition_edge_callback(ENTER, interface_state.ConfigureSettings)
    def enter_ConfigureSettings(self, old_state, new_state):
        self.show_component('settingsPage', True)

    @transition_edge_callback(EXIT, interface_state.ConfigureSettings)
    def exit_ConfigureSettings(self, old_state, new_state):
        self.show_component('settingsPage', False)

    @transition_edge_callback(ENTER, interface_state.configure_server_states)
    def enter_search(self, old_state, new_state):
        if not isinstance(old_state, interface_state.configure_server_states):
            self.builder.get_object('findYourInstituteSearch').grab_focus()
            search.show_result_components(self.builder, True)
            search.show_search_components(self.builder, True)
            search.init_server_search(self.builder)
            search.connect_selection_handlers(
                self.builder, self.on_select_server)

    @transition_edge_callback(EXIT, interface_state.configure_server_states)
    def exit_search(self, old_state, new_state):
        if not isinstance(new_state, interface_state.configure_server_states):
            search.show_result_components(self.builder, False)
            search.show_search_components(self.builder, False)
            search.exit_server_search(self.builder)
            search.disconnect_selection_handlers(
                self.builder, self.on_select_server)
            self.set_search_text('')

    @transition_edge_callback(
        ENTER, interface_state.PendingConfigurePredefinedServer)
    def enter_PendingConfigurePredefinedServer(self, old_state, new_state):
        search.update_results(self.builder, [])
        if not isinstance(old_state, interface_state.configure_server_states):
            self.set_search_text(new_state.search_query)

    @transition_edge_callback(ENTER, interface_state.ConfigurePredefinedServer)
    def enter_ConfigurePredefinedServer(self, old_state, new_state):
        search.update_results(self.builder, new_state.results)
        if not isinstance(old_state, interface_state.configure_server_states):
            self.set_search_text(new_state.search_query)

    @transition_edge_callback(ENTER, interface_state.ConfigureCustomServer)
    def enter_ConfigureCustomServer(self, old_state, new_state):
        search.update_results(self.builder, [CustomServer(new_state.address)])
        if not isinstance(old_state, interface_state.configure_server_states):
            self.set_search_text(new_state.address)

    @transition_edge_callback(ENTER, interface_state.MainState)
    def enter_MainState(self, old_state, new_state):
        search.show_result_components(self.builder, True)
        self.show_component('addOtherServerRow', True)
        self.show_component('addOtherServerButton', True)
        search.update_results(self.builder, new_state.servers)
        search.init_server_search(self.builder)
        search.connect_selection_handlers(
            self.builder, self.on_select_server)

    @transition_edge_callback(EXIT, interface_state.MainState)
    def exit_MainState(self, old_state, new_state):
        search.show_result_components(self.builder, False)
        self.show_component('addOtherServerRow', False)
        self.show_component('addOtherServerButton', False)
        search.exit_server_search(self.builder)
        search.disconnect_selection_handlers(
            self.builder, self.on_select_server)

    @transition_edge_callback(ENTER, interface_state.OAuthSetupPending)
    @transition_edge_callback(ENTER, interface_state.OAuthSetup)
    def enter_oauth_setup(self, old_state, new_state):
        self.show_component('openBrowserPage', True)
        self.show_component('cancelBrowserButton',
                            isinstance(new_state, interface_state.OAuthSetup))

    @transition_edge_callback(EXIT, interface_state.OAuthSetupPending)
    @transition_edge_callback(EXIT, interface_state.OAuthSetup)
    def exit_oauth_setup(self, old_state, new_state):
        self.show_component('openBrowserPage', False)
        self.show_component('cancelBrowserButton', False)

    @transition_edge_callback(ENTER, interface_state.OAuthRefreshToken)
    def enter_OAuthRefreshToken(self, old_state, new_state):
        self.show_message_page(
            _("Finishing Authorization"),
            _("The authorization token is being finished."),
        )

    @transition_edge_callback(EXIT, interface_state.OAuthRefreshToken)
    def exit_OAuthRefreshToken(self, old_state, new_state):
        self.hide_message_page()

    @transition_edge_callback(ENTER, interface_state.LoadingServerInformation)
    def enter_LoadingServerInformation(self, old_state, new_state):
        self.show_message_page(
            _("Loading"),
            _("The server details are being loaded."),
        )

    @transition_edge_callback(EXIT, interface_state.LoadingServerInformation)
    def exit_LoadingServerInformation(self, old_state, new_state):
        self.hide_message_page()

    @transition_edge_callback(ENTER, interface_state.ChooseProfile)
    def enter_ChooseProfile(self, old_state, new_state):
        self.show_component('chooseProfilePage', True)
        self.show_component('profileTreeView', True)

        profile_tree_view = self.builder.get_object('profileTreeView')
        profiles_list_model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_PYOBJECT)

        if len(profile_tree_view.get_columns()) == 0:
            # Only initialize this tree view once.
            text_cell = Gtk.CellRendererText()
            text_cell.set_property("size-points", 14)

            column = Gtk.TreeViewColumn(None, text_cell, text=0)
            profile_tree_view.append_column(column)

            profile_tree_view.set_model(profiles_list_model)

        selection = profile_tree_view.get_selection()
        selection.connect("changed", self.on_profile_selection_changed)

        profiles_list_model.clear()
        for profile in new_state.profiles:
            profiles_list_model.append([str(profile), profile])

    @transition_edge_callback(EXIT, interface_state.ChooseProfile)
    def exit_ChooseProfile(self, old_state, new_state):
        self.show_component('chooseProfilePage', False)
        self.show_component('profileTreeView', False)

        profile_tree_view = self.builder.get_object('profileTreeView')
        selection = profile_tree_view.get_selection()
        selection.disconnect_by_func(self.on_profile_selection_changed)

    @transition_edge_callback(ENTER, interface_state.ChooseSecureInternetLocation)
    def enter_ChooseSecureInternetLocation(self, old_state, new_state):
        self.show_component('chooseLocationPage', True)
        self.show_component('locationTreeView', True)

        location_tree_view = self.builder.get_object('locationTreeView')
        location_list_model = Gtk.ListStore(GObject.TYPE_STRING, GdkPixbuf.Pixbuf, GObject.TYPE_PYOBJECT)

        if len(location_tree_view.get_columns()) == 0:
            # Only initialize this tree view once.
            text_cell = Gtk.CellRendererText()
            text_cell.set_property("size-points", 14)

            renderer_pixbuf = Gtk.CellRendererPixbuf()
            column = Gtk.TreeViewColumn("Image", renderer_pixbuf, pixbuf=1)
            location_tree_view.append_column(column)

            column = Gtk.TreeViewColumn(None, text_cell, text=0)
            location_tree_view.append_column(column)

            location_tree_view.set_model(location_list_model)

        selection = location_tree_view.get_selection()
        selection.connect("changed", self.on_location_selection_changed)

        location_list_model.clear()
        for location in new_state.locations:
            if location.flag_path is None:
                continue  # TODO
            else:
                flag = GdkPixbuf.Pixbuf.new_from_file(location.flag_path)
            location_list_model.append([location.country_name, flag, location])

    @transition_edge_callback(EXIT, interface_state.ChooseSecureInternetLocation)
    def exit_ChooseSecureInternetLocation(self, old_state, new_state):
        self.show_component('chooseLocationPage', False)
        self.show_component('locationTreeView', False)

        location_tree_view = self.builder.get_object('locationTreeView')
        selection = location_tree_view.get_selection()
        selection.disconnect_by_func(self.on_location_selection_changed)

    @transition_edge_callback(ENTER, interface_state.ConfiguringConnection)
    def enter_ConfiguringConnection(self, old_state, new_state):
        self.show_message_page(
            _("Configuring"),
            _("Your connection is being configured."),
        )

    @transition_edge_callback(EXIT, interface_state.ConfiguringConnection)
    def exit_ConfiguringConnection(self, old_state, new_state):
        self.hide_message_page()

    @transition_edge_callback(ENTER, interface_state.ConnectionStatus)
    def enter_ConnectionStatus(self, old_state, new_state):
        self.show_component('connectionPage', True)
        self.show_component('serverLabel', True)
        self.show_component('connectionStatusImage', True)
        self.show_component('connectionStatusLabel', True)
        self.show_component('connectionSubStatus', True)

        self.update_connection_server()
        self.update_connection_status()

        if hasattr(self, '_cancel_validity_updates'):
            # Cancel any previous threads, as they might have been cancelled
            # and there shouln't be multiple threads running.
            self._cancel_validity_updates()

        def update_connection_validity():
            # This function runs in a background thread.
            state = self.app.interface_state
            if not isinstance(state, interface_state.ConnectionStatus):
                # cancel this thread
                return False
            run_in_main_gtk_thread(self.update_connection_validity)()
            if datetime.utcnow() < state.validity.end:
                return True
            else:
                self.app.network_transition_threadsafe('set_certificate_expired')
                return False

        self._cancel_validity_updates = run_periodically(
            update_connection_validity,
            UPDATE_EXIPRY_INTERVAL,
            'update-validity',
        )

    @transition_edge_callback(EXIT, interface_state.ConnectionStatus)
    def exit_ConnectionStatus(self, old_state, new_state):
        self.show_component('connectionPage', False)

        if hasattr(self, '_cancel_validity_updates'):
            self._cancel_validity_updates()
            del self._cancel_validity_updates

    @transition_edge_callback(ENTER, interface_state.ErrorState)
    def enter_ErrorState(self, old_state, new_state):
        self.show_component('messagePage', True)
        self.show_component('messageLabel', True)
        self.show_component('messageText', True)
        self.set_text('messageLabel', _("Error"))
        self.set_text('messageText', new_state.message)
        self.show_component('messageButton', True)
        self.builder.get_object('messageButton').set_label(_("Ok"))
        button = self.builder.get_object('messageButton')
        button.connect("clicked", self.on_acknowledge_error)

    @transition_edge_callback(EXIT, interface_state.ErrorState)
    def exit_ErrorState(self, old_state, new_state):
        self.show_component('messagePage', False)
        button = self.builder.get_object('messageButton')
        button.disconnect_by_func(self.on_acknowledge_error)

    # ui callbacks

    def on_settings_button_released(self, widget, event):
        logger.debug("clicked settings button")
        self.app.interface_transition('toggle_settings')

    def on_help_button_released(self, widget, event):
        logger.debug("clicked help button")
        webbrowser.open(HELP_URL)

    def on_back_button_released(self, widget, event):
        logger.debug("clicked back button")
        self.app.interface_transition('go_back')

    def on_add_other_server_button_clicked(self, button) -> None:
        logger.debug("on_add_other_server_button_clicked")
        self.app.interface_transition('configure_new_server')

    def on_select_server(self, selection):
        logger.debug("selected search result")
        (model, tree_iter) = selection.get_selected()
        selection.unselect_all()
        if tree_iter is None:
            logger.info("selection empty")
        else:
            row = model[tree_iter]
            server = row[1]
            logger.debug(f"selected server: {server!r}")
            self.app.interface_transition('connect_to_server', server)

    def on_cancel_oauth_setup_button_clicked(self, _):
        self.app.interface_transition('oauth_setup_cancel')

    def on_search_changed(self, _=None):
        query = self.builder.get_object('findYourInstituteSearch').get_text()
        logger.debug(f"entered search query: {query}")
        if self.lets_connect or query.count('.') >= 2:
            # Anything with two periods is interpreted
            # as a custom server address.
            self.app.interface_transition(
                'enter_custom_address', address=query)
        else:
            self.app.interface_transition(
                'enter_search_query', search_query=query)

    def on_activate_changed(self, _=None):
        logger.debug("on_activate_changed")
        # TODO

    def on_connection_switch_state_set(self, switch, state):
        logger.debug("on_activate_changed")
        if state:
            self.app.interface_transition('activate_connection')
        else:
            self.app.interface_transition('deactivate_connection')
        return True

    def on_profile_selection_changed(self, selection):
        logger.debug("selected profile")
        (model, tree_iter) = selection.get_selected()
        selection.unselect_all()
        if tree_iter is None:
            logger.info("selection empty")
        else:
            row = model[tree_iter]
            profile = row[1]
            logger.debug(f"selected profile: {profile!r}")
            self.app.interface_transition('select_profile', profile)

    def on_location_selection_changed(self, selection):
        logger.debug("selected location")
        (model, tree_iter) = selection.get_selected()
        selection.unselect_all()
        if tree_iter is None:
            logger.info("selection empty")
        else:
            row = model[tree_iter]
            location = row[2]
            logger.debug(f"selected location: {location!r}")
            self.app.interface_transition('select_secure_internet_location', location)

    def on_acknowledge_error(self, event):
        self.app.interface_transition('acknowledge_error')

    def on_renew_session_clicked(self, event):
        self.app.network_transition('renew_certificate')



from typing import Sequence, Optional
import logging
import traceback
import signal
from argparse import ArgumentParser
from os import geteuid
from sys import exit, argv
import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
from gi.repository import Gtk
from eduvpn import __version__

logger = logging.getLogger(__name__)
log_format = format_ = (
    '%(asctime)s - %(threadName)s - %(levelname)s - %(name)s'
    ' - %(filename)s:%(lineno)d - %(message)s'
)


def parse_args(args: Optional[Sequence[str]]) -> int:
    """
    Parses command line arguments:
    returns:
        logging_level
    """
    parser = ArgumentParser()
    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help="enable debug logging",
    )
    parser.add_argument(
        '-v',
        '--version',
        action='store_true',
        help="print version and exit",
    )
    parsed = parser.parse_args(args=args)

    if parsed.version:
        print("eduVPN Linux client version {}".format(__version__))
        exit(0)

    if parsed.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    return level


def signal_handler(*_, **__):
    exit(0)


def main_loop(args=None, lets_connect=False):
    if args is None:
        args = argv[1:]
    loglevel = parse_args(args)
    logging.basicConfig(level=loglevel, format=log_format)

    signal.signal(signal.SIGINT, signal_handler)

    if geteuid() == 0:
        logger.error("Running client as root is not supported (yet)")
        exit(1)

    try:
        # import this later so the logging is properly configured
        from eduvpn.ui.ui import EduVpnGui

        edu_vpn_gui = EduVpnGui(lets_connect)
        edu_vpn_gui.run()
    except Exception as e:
        fatal_reason = f"Caught exception: {e}"
        logger.error(fatal_reason)
        logging.error(traceback.format_exc())
        dialog = Gtk.MessageDialog(flags=Gtk.DialogFlags.MODAL,
                                   type=Gtk.MessageType.ERROR,
                                   buttons=Gtk.ButtonsType.OK,
                                   message_format=fatal_reason)
        dialog.connect("response", Gtk.main_quit)
        dialog.show()

    Gtk.main()


def eduvpn(args=None):
    main_loop(args)


def letsconnect(args=None):
    main_loop(args, lets_connect=True)


if __name__ == '__main__':
    eduvpn()



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Set
import unittest
from unittest.mock import Mock

from eduvpn.state_machine import (
    ENTER, EXIT, StateMachine, transition_callback, transition_edge_callback,
    InvalidStateTransition)
from tests.mock_config import mock_config_dict


class StateMachineTests(unittest.TestCase):
    def test_state_transition(self):
        initial_state = Mock()
        final_state = object()
        initial_state.perform_the_transition.return_value = final_state

        sm = StateMachine(initial_state)
        self.assertIs(sm.state, initial_state)

        sm.transition('perform_the_transition', 1, x=2)
        self.assertIs(sm.state, final_state)
        initial_state.perform_the_transition.assert_called_once_with(1, x=2)

    def test_generic_callback(self):
        initial_state = Mock()
        final_state = object()
        initial_state.perform_the_transition.return_value = final_state
        callback = Mock()

        sm = StateMachine(initial_state)
        sm.register_generic_callback(callback)
        sm.transition('perform_the_transition')
        callback.assert_called_once_with(initial_state, final_state)

    def test_edge_callback(self):
        class InitialState:
            def perform_the_transition(self):
                return final_state

        class FinalState:
            pass

        class OtherState:
            pass

        initial_state = InitialState()
        final_state = FinalState()

        enter_initial_callback = Mock()
        exit_initial_callback = Mock()
        enter_final_callback = Mock()
        exit_final_callback = Mock()
        other_callback = Mock()

        sm = StateMachine(initial_state)
        sm.register_edge_callback(InitialState, ENTER, enter_initial_callback)
        sm.register_edge_callback(InitialState, EXIT, exit_initial_callback)
        sm.register_edge_callback(FinalState, ENTER, enter_final_callback)
        sm.register_edge_callback(FinalState, EXIT, exit_final_callback)
        sm.register_edge_callback(OtherState, ENTER, other_callback)
        sm.register_edge_callback(OtherState, EXIT, other_callback)

        sm.transition('perform_the_transition')
        exit_initial_callback.assert_called_once_with(
            initial_state, final_state)
        enter_final_callback.assert_called_once_with(
            initial_state, final_state)

        enter_initial_callback.assert_not_called()
        exit_final_callback.assert_not_called()
        other_callback.assert_not_called()

    def test_connect_object_callbacks(self):
        class BaseState:
            pass

        class InitialState(BaseState):
            def perform_the_transition(self):
                return final_state

        class FinalState(BaseState):
            pass

        class OtherBaseState:
            pass

        initial_state = InitialState()
        final_state = FinalState()

        class Connector:
            def __init__(self):
                self.calls: Set[str] = set()

            @transition_callback(BaseState)
            def any(self, old, new):
                self.calls.add(('any', old, new))

            @transition_edge_callback(ENTER, InitialState)
            def enter_initial(self, old, new):
                self.calls.add(('enter_initial', old, new))

            @transition_edge_callback(EXIT, InitialState)
            def exit_initial(self, old, new):
                self.calls.add(('exit_initial', old, new))

            @transition_callback(OtherBaseState)
            def other(self, old, new):
                # This callback targets another state machine.
                raise AssertionError

        connector = Connector()
        sm = StateMachine(initial_state)
        sm.connect_object_callbacks(connector, BaseState)
        sm.transition('perform_the_transition')
        self.assertEqual(connector.calls, {
            ('any', initial_state, final_state),
            ('exit_initial', initial_state, final_state),
        })
